import abc
import pathlib
from typing import Optional

from robotoff import settings
from robotoff.insights import InsightType
from robotoff.models import ProductInsight
from robotoff.mongo import MONGO_CLIENT_CACHE
from robotoff.off import generate_image_url
from robotoff.products import get_product
from robotoff.taxonomy import Taxonomy, TaxonomyType, get_taxonomy
from robotoff.utils import get_logger, load_json
from robotoff.utils.i18n import TranslationStore
from robotoff.utils.types import JSONType

logger = get_logger(__name__)

DISPLAY_IMAGE_SIZE = 400
SMALL_IMAGE_SIZE = 200
THUMB_IMAGE_SIZE = 100


LABEL_IMAGES: dict[str, str] = load_json(settings.LABEL_LOGOS_PATH)  # type: ignore


class Question(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def serialize(self) -> JSONType:
        pass

    @abc.abstractmethod
    def get_type(self):
        pass


class AddBinaryQuestion(Question):
    """JSON formatter for a binary (yes/no) question.

    :param question: The non-localized question text (in English)
    :param value: the suggested answer to the question
    :param insight: the insight used to generate the question
    :param ref_image_url: the URL of a reference image, used to display the
    image of the logo for `label` insights, optional
    :param source_image_url: the URL of the image from where the insight was
    extracted from, optional
    :param value_tag: The `value_tag` that is going to be sent to Product
    Opener if the insight is validated, optional
    """

    def __init__(
        self,
        question: str,
        value: str,
        insight: ProductInsight,
        ref_image_url: Optional[str] = None,
        source_image_url: Optional[str] = None,
        value_tag: Optional[str] = None,
    ):
        self.question: str = question
        self.value: str = value
        self.insight_id: str = str(insight.id)
        self.insight_type: str = str(insight.type)
        self.barcode: str = insight.barcode
        self.ref_image_url: Optional[str] = ref_image_url
        self.source_image_url: Optional[str] = source_image_url
        self.value_tag: Optional[str] = value_tag

    def get_type(self):
        return "add-binary"

    def serialize(self) -> JSONType:
        serial = {
            "barcode": self.barcode,
            "type": self.get_type(),
            "value": self.value,
            "question": self.question,
            "insight_id": self.insight_id,
            "insight_type": self.insight_type,
        }

        if self.value_tag:
            serial["value_tag"] = self.value_tag

        if self.ref_image_url:
            serial["ref_image_url"] = self.ref_image_url

        if self.source_image_url:
            serial["source_image_url"] = self.source_image_url

        return serial


class IngredientSpellcheckQuestion(Question):
    def __init__(self, insight: ProductInsight, ref_image_url: Optional[str]):
        self.insight_id: str = str(insight.id)
        self.insight_type: str = str(insight.type)
        self.barcode: str = insight.barcode
        self.corrected: str = insight.data["corrected"]
        self.text: str = insight.data["text"]
        self.corrections: list[JSONType] = insight.data["corrections"]
        self.lang: str = insight.data["lang"]
        self.ref_image_url: Optional[str] = ref_image_url

    def get_type(self):
        return "ingredient-spellcheck"

    def serialize(self) -> JSONType:
        serial = {
            "barcode": self.barcode,
            "type": self.get_type(),
            "insight_id": self.insight_id,
            "insight_type": self.insight_type,
            "text": self.text,
            "corrected": self.corrected,
            "corrections": self.corrections,
            "lang": self.lang,
        }

        if self.ref_image_url:
            serial["ref_image_url"] = self.ref_image_url

        return serial


class QuestionFormatter(metaclass=abc.ABCMeta):
    def __init__(self, translation_store: TranslationStore):
        self.translation_store: TranslationStore = translation_store

    @abc.abstractmethod
    def format_question(self, insight: ProductInsight, lang: str) -> Question:
        pass


class CategoryQuestionFormatter(QuestionFormatter):
    question = "Does the product belong to this category?"

    def format_question(self, insight: ProductInsight, lang: str) -> Question:
        taxonomy: Taxonomy = get_taxonomy(TaxonomyType.category.name)
        localized_value: str = taxonomy.get_localized_name(insight.value_tag, lang)
        localized_question = self.translation_store.gettext(lang, self.question)
        source_image_url = self.get_source_image_url(insight.barcode)
        return AddBinaryQuestion(
            question=localized_question,
            value=localized_value,
            value_tag=insight.value_tag,
            insight=insight,
            source_image_url=source_image_url,
        )

    @staticmethod
    def get_source_image_url(barcode: str) -> Optional[str]:
        product: Optional[JSONType] = get_product(barcode)

        if product is None or "images" not in product:
            return None

        selected_images = CategoryQuestionFormatter.generate_selected_images(
            product["images"], barcode
        )

        for key in ("front", "ingredients", "nutrition"):
            if key in selected_images:
                images = selected_images[key]

                if "display" in images:
                    display_images = list(images["display"].values())

                    if display_images:
                        return display_images[0]

        return None

    @staticmethod
    def generate_selected_images(
        images: JSONType, barcode: str
    ) -> dict[str, dict[str, dict[str, str]]]:
        """Generate the same `selected_images` field as returned by Product
        Opener API.

        :param images: the `images` data of the product
        :param barcode: the product barcode
        :return: the `selected_images` data
        """
        selected_images: dict[str, dict[str, dict[str, str]]] = {
            image_type: {}
            for image_type in ("front", "nutrition", "ingredients", "packaging")
        }

        for key, image_data in images.items():
            if (
                key.startswith("front_")
                or key.startswith("nutrition_")
                or key.startswith("ingredients_")
                or key.startswith("packaging_")
            ):
                image_type = key.split("_")[
                    0
                ]  # to get image type: `front`, `nutrition`, `ingredients` or `packaging`
                language = key.split("_")[1]  # splitting to get the language name
                revision_id = image_data["rev"]  # get revision_id for all languages
                available_image_sizes = set(
                    int(size) for size in image_data["sizes"] if size.isdigit()
                )

                for field_name, image_size in (
                    ("display", DISPLAY_IMAGE_SIZE),
                    ("small", SMALL_IMAGE_SIZE),
                    ("thumb", THUMB_IMAGE_SIZE),
                ):
                    if image_size in available_image_sizes:
                        image_url = generate_image_url(
                            barcode, f"{key}.{revision_id}.{image_size}"
                        )
                        selected_images[image_type].setdefault(field_name, {})
                        selected_images[image_type][field_name][language] = image_url

        return selected_images


class ProductWeightQuestionFormatter(QuestionFormatter):
    question = "Does this weight match the weight displayed on the product?"

    def format_question(self, insight: ProductInsight, lang: str) -> Question:
        localized_question = self.translation_store.gettext(lang, self.question)

        source_image_url = None
        if insight.source_image:
            source_image_url = settings.OFF_IMAGE_BASE_URL + get_display_image(
                insight.source_image
            )

        return AddBinaryQuestion(
            question=localized_question,
            value=insight.value,
            insight=insight,
            source_image_url=source_image_url,
        )


class LabelQuestionFormatter(QuestionFormatter):
    question = "Does the product have this label?"

    def format_question(self, insight: ProductInsight, lang: str) -> Question:
        value_tag: str = insight.value_tag
        ref_image_url = LABEL_IMAGES.get(value_tag)

        taxonomy: Taxonomy = get_taxonomy(TaxonomyType.label.name)
        localized_value: str = taxonomy.get_localized_name(value_tag, lang)
        localized_question = self.translation_store.gettext(lang, self.question)

        source_image_url = None
        if insight.source_image:
            source_image_url = settings.OFF_IMAGE_BASE_URL + get_display_image(
                insight.source_image
            )

        return AddBinaryQuestion(
            question=localized_question,
            value=localized_value,
            value_tag=value_tag,
            insight=insight,
            ref_image_url=ref_image_url,
            source_image_url=source_image_url,
        )


class BrandQuestionFormatter(QuestionFormatter):
    question = "Does the product belong to this brand?"

    def format_question(self, insight: ProductInsight, lang: str) -> Question:
        localized_question = self.translation_store.gettext(lang, self.question)

        source_image_url = None
        if insight.source_image:
            source_image_url = settings.OFF_IMAGE_BASE_URL + get_display_image(
                insight.source_image
            )

        return AddBinaryQuestion(
            question=localized_question,
            value=insight.value,
            value_tag=insight.value_tag,
            insight=insight,
            source_image_url=source_image_url,
        )


class IngredientSpellcheckQuestionFormatter(QuestionFormatter):
    def format_question(self, insight: ProductInsight, lang: str) -> Question:
        ref_image_url = self.get_ingredient_image_url(insight.barcode, lang)
        return IngredientSpellcheckQuestion(
            insight=insight, ref_image_url=ref_image_url
        )

    def get_ingredient_image_url(self, barcode: str, lang: str) -> Optional[str]:
        mongo_client = MONGO_CLIENT_CACHE.get()
        collection = mongo_client.off.products
        product = collection.find_one({"code": barcode}, ["images"])

        if product is None:
            return None

        images = product.get("images", {})
        field_name = "ingredients_{}".format(lang)

        if field_name in images:
            image = images[field_name]
            image_id = "ingredients_{}.{}.full".format(lang, image["rev"])
            return generate_image_url(barcode, image_id)

        return None


class NutritionImageQuestionFormatter(QuestionFormatter):
    question = "Is this image a nutrition image for this language?"

    def format_question(self, insight: ProductInsight, lang: str) -> Question:
        localized_question = self.translation_store.gettext(lang, self.question)

        source_image_url = None
        if insight.source_image:
            source_image_url = settings.OFF_IMAGE_BASE_URL + get_display_image(
                insight.source_image
            )

        return AddBinaryQuestion(
            question=localized_question,
            value=insight.value_tag,
            insight=insight,
            source_image_url=source_image_url,
        )


def get_display_image(source_image: str) -> str:
    image_path = pathlib.Path(source_image)

    if not image_path.stem.isdigit():
        return source_image

    display_name = "{}.400.jpg".format(image_path.name.split(".")[0])
    return str(image_path.parent / display_name)


class QuestionFormatterFactory:
    formatters: dict[str, type] = {
        InsightType.category.name: CategoryQuestionFormatter,
        InsightType.label.name: LabelQuestionFormatter,
        InsightType.product_weight.name: ProductWeightQuestionFormatter,
        InsightType.brand.name: BrandQuestionFormatter,
        InsightType.ingredient_spellcheck.name: IngredientSpellcheckQuestionFormatter,
        InsightType.nutrition_image.name: NutritionImageQuestionFormatter,
    }

    @classmethod
    def get(cls, insight_type: str):
        return cls.formatters.get(insight_type)

    @classmethod
    def get_available_types(cls) -> list[str]:
        return list(cls.formatters.keys())

    @classmethod
    def get_default_types(cls) -> list[str]:
        return [
            InsightType.category.name,
            InsightType.label.name,
            InsightType.product_weight.name,
            InsightType.brand.name,
        ]



import abc

from robotoff.spellcheck.items import SpellcheckItem


class BaseSpellchecker(metaclass=abc.ABCMeta):
    def predict(self, items: list[SpellcheckItem]) -> list[SpellcheckItem]:
        return [self.predict_one(item) for item in items]

    def predict_one(self, item: SpellcheckItem) -> SpellcheckItem:
        if item.is_lang_allowed:
            item.update_correction(
                correction=self.correct(item.latest_correction), model=self.name
            )
        return item

    @abc.abstractmethod
    def correct(self, text: str) -> str:
        pass

    @abc.abstractmethod
    def get_config(self):
        pass

    @property
    def name(self) -> str:
        return self.__class__.__name__



from typing import Optional, Union

from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType

from .dataclass import OCRResult


def get_image_lang(ocr_result: Union[OCRResult, str]) -> list[Prediction]:
    if isinstance(ocr_result, str):
        return []

    image_lang: Optional[dict[str, int]] = ocr_result.get_languages()

    if image_lang is None:
        return []

    words = image_lang["words"]
    percents = {}
    for key, count in image_lang.items():
        if key == "words":
            continue

        percents[key] = count * 100 / words

    return [
        Prediction(
            type=PredictionType.image_lang,
            data={"count": image_lang, "percent": percents},
        )
    ]



import functools
from typing import Iterable, Optional, Union

from flashtext import KeywordProcessor

from robotoff import settings
from robotoff.brands import BRAND_BLACKLIST_STORE, keep_brand_from_taxonomy
from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType
from robotoff.utils import get_logger, text_file_iter
from robotoff.utils.text import get_tag

from .dataclass import OCRResult, get_text
from .utils import generate_keyword_processor

logger = get_logger(__name__)


def generate_brand_keyword_processor(
    brands: Iterable[str],
    blacklist: bool = True,
):
    blacklisted_brands: Optional[set[str]] = None
    if blacklist:
        blacklisted_brands = BRAND_BLACKLIST_STORE.get()

    keep_func = functools.partial(
        keep_brand_from_taxonomy,
        blacklisted_brands=blacklisted_brands,
    )
    return generate_keyword_processor(brands, keep_func=keep_func)


def get_logo_annotation_brands() -> dict[str, str]:
    brands: dict[str, str] = {}

    for item in text_file_iter(settings.OCR_LOGO_ANNOTATION_BRANDS_DATA_PATH):
        if "||" in item:
            logo_description, label_tag = item.split("||")
        else:
            logger.warning("'||' separator expected!")
            continue

        brands[logo_description] = label_tag

    return brands


LOGO_ANNOTATION_BRANDS: dict[str, str] = get_logo_annotation_brands()
TAXONOMY_BRAND_PROCESSOR = generate_brand_keyword_processor(
    text_file_iter(settings.OCR_TAXONOMY_BRANDS_PATH)
)
BRAND_PROCESSOR = generate_brand_keyword_processor(
    text_file_iter(settings.OCR_BRANDS_PATH),
)


def extract_brands(
    processor: KeywordProcessor,
    text: str,
    data_source_name: str,
    automatic_processing: bool,
) -> list[Prediction]:
    predictions = []

    for (brand_tag, brand), span_start, span_end in processor.extract_keywords(
        text, span_info=True
    ):
        match_str = text[span_start:span_end]
        predictions.append(
            Prediction(
                type=PredictionType.brand,
                value=brand,
                value_tag=brand_tag,
                automatic_processing=automatic_processing,
                predictor=data_source_name,
                data={"text": match_str, "notify": False},
            )
        )

    return predictions


def extract_brands_google_cloud_vision(ocr_result: OCRResult) -> list[Prediction]:
    predictions = []
    for logo_annotation in ocr_result.logo_annotations:
        if logo_annotation.description in LOGO_ANNOTATION_BRANDS:
            brand = LOGO_ANNOTATION_BRANDS[logo_annotation.description]

            predictions.append(
                Prediction(
                    type=PredictionType.brand,
                    value=brand,
                    value_tag=get_tag(brand),
                    automatic_processing=False,
                    predictor="google-cloud-vision",
                    data={"confidence": logo_annotation.score, "notify": False},
                )
            )

    return predictions


def find_brands(content: Union[OCRResult, str]) -> list[Prediction]:
    predictions: list[Prediction] = []
    text = get_text(content)

    if text:
        predictions += extract_brands(
            BRAND_PROCESSOR, text, "curated-list", automatic_processing=True
        )
        predictions += extract_brands(
            TAXONOMY_BRAND_PROCESSOR, text, "taxonomy", automatic_processing=False
        )

    if isinstance(content, OCRResult):
        predictions += extract_brands_google_cloud_vision(content)

    return predictions



from pathlib import Path
from typing import Iterable

import orjson
from elasticsearch import Elasticsearch

from robotoff import settings
from robotoff.elasticsearch.product.dump import generate_product_data
from robotoff.utils import get_logger
from robotoff.utils.es import perform_export

logger = get_logger(__name__)


class ElasticsearchExporter:
    """ElasticsearchExporter exports new index data to Elasticsearch."""

    def __init__(self, es_client: Elasticsearch):
        self.es_client = es_client

    def _delete_existing_data(self, index: str) -> None:
        resp = self.es_client.delete_by_query(
            body={"query": {"match_all": {}}},
            index=index,
            ignore_unavailable=True,
            doc_type=settings.ELASTICSEARCH_TYPE,
        )

        logger.info("Deleted %d documents from %s", resp["deleted"], index)

    def _get_data(self, index: str) -> Iterable[tuple[str, dict]]:
        if index == settings.ElasticsearchIndex.PRODUCT:
            return generate_product_data()

        raise ValueError(f"unknown index: {index}")

    def load_index(self, index: str, index_filepath: Path) -> None:
        """Creates the given index if it doesn't already exist."""
        if not self.es_client.indices.exists(index=index):
            logger.info("Creating index: %s", index)
            with open(index_filepath, "rb") as f:
                conf = orjson.loads(f.read())
            self.es_client.indices.create(index=index, body=conf)

    def export_index_data(self, index: str) -> int:
        """Given the index to export data for, this function removes existing data and exports a newer version.

        .. warning: right now, we delete then recreate the index.
           This means that as this method runs,
           some request might be silently handled erroneously (with a partial index).
           This is not a problem right now, as we don't have *real-time* requests,
           but only async ones for categories.

        Returns the number of rows inserted into the index."""
        logger.info("Deleting existing %s data...", index)
        self._delete_existing_data(index)

        index_data = self._get_data(index)

        logger.info("Starting %s export to Elasticsearch...", index)

        rows_inserted = perform_export(self.es_client, index_data, index)

        logger.info("Inserted %d rows for index %s", rows_inserted, index)
        return rows_inserted



import sys

from rq import Connection, Worker

from robotoff import settings
from robotoff.models import with_db
from robotoff.utils import get_logger
from robotoff.workers.queues import redis_conn

logger = get_logger()
settings.init_sentry()


@with_db
def load_resources(refresh: bool = False):
    """Load cacheable resources in memory.

    This way, all resources are available in memory before the worker forks.
    """
    if refresh:
        logger.info("Refreshing worker resource caches...")
    else:
        logger.info("Loading resources in memory...")

    from robotoff import logos, taxonomy
    from robotoff.prediction.category import matcher
    from robotoff.prediction.object_detection import ObjectDetectionModelRegistry

    matcher.load_resources()
    taxonomy.load_resources()
    logos.load_resources()

    if not refresh:
        logger.info("Loading object detection model labels...")
        ObjectDetectionModelRegistry.load_all()


class CustomWorker(Worker):
    def run_maintenance_tasks(self):
        super().run_maintenance_tasks()
        load_resources(refresh=True)


def run(queues: list[str], burst: bool = False):
    load_resources()
    try:
        with Connection(connection=redis_conn):
            w = CustomWorker(queues=queues)
            w.work(logging_level="INFO", burst=burst)
    except ConnectionError as e:
        print(e)
        sys.exit(1)



from robotoff.models import db


class DBConnectionMiddleware:
    def process_resource(self, req, resp, resource, params):
        db.connect(reuse_if_open=True)

    def process_response(self, req, resp, resource, req_succeeded):
        if not db.is_closed():
            db.close()



from typing import Iterable, Optional, Union

from robotoff import settings
from robotoff.prediction.types import Prediction
from robotoff.products import ProductDataset
from robotoff.spellcheck.base_spellchecker import BaseSpellchecker
from robotoff.spellcheck.elasticsearch import ElasticSearchSpellchecker
from robotoff.spellcheck.items import SpellcheckItem
from robotoff.spellcheck.patterns import PatternsSpellchecker
from robotoff.spellcheck.percentages import PercentagesSpellchecker
from robotoff.spellcheck.vocabulary import VocabularySpellchecker
from robotoff.types import PredictionType
from robotoff.utils.types import JSONType

SPELLCHECKERS = {
    "elasticsearch": ElasticSearchSpellchecker,
    "patterns": PatternsSpellchecker,
    "percentages": PercentagesSpellchecker,
    "vocabulary": VocabularySpellchecker,
}


class Spellchecker:
    def __init__(self, pipeline: list[BaseSpellchecker]):
        self.spellcheckers: list[BaseSpellchecker] = pipeline

    @classmethod
    def load(
        cls,
        client,
        pipeline: Optional[list[Union[str, BaseSpellchecker]]] = None,
        **es_kwargs,
    ):
        pipeline_: list[BaseSpellchecker] = []

        if pipeline is None:
            pipeline = ["patterns", "elasticsearch", "vocabulary"]

        for item in pipeline:
            if isinstance(item, str):
                if item not in SPELLCHECKERS:
                    raise ValueError(
                        f"Spellchecker {item} not found. Available : {list(SPELLCHECKERS.keys())}"
                    )

                if item == "elasticsearch":
                    base_spellchecker = ElasticSearchSpellchecker(client, **es_kwargs)
                else:
                    base_spellchecker = SPELLCHECKERS[item]()

                pipeline_.append(base_spellchecker)

            elif not isinstance(item, BaseSpellchecker):
                raise TypeError(
                    f"invalid item in pipeline: {item}, expected str or BaseSpellchecker"
                )
                pipeline_.append(item)

        return cls(pipeline_)

    def generate_insights(
        self,
        max_errors: Optional[int] = None,
        lang: str = "fr",
        limit: Optional[int] = None,
    ) -> Iterable[Prediction]:
        dataset = ProductDataset(settings.JSONL_DATASET_PATH)
        product_iter = (
            dataset.stream()
            .filter_by_country_tag("en:france")
            .filter_text_field("lang", lang)
            .filter_nonempty_text_field("ingredients_text_fr")
            .iter()
        )

        insights_count = 0
        for product in product_iter:
            if self.is_product_valid(product, max_errors=max_errors):
                insight = self.predict_insight(product["ingredients_text_fr"])
                if insight is not None:
                    insight["lang"] = lang
                    yield Prediction(
                        type=PredictionType.ingredient_spellcheck,
                        data=insight,
                        barcode=product["code"],
                    )

                    insights_count += 1
                    if limit is not None and insights_count >= limit:
                        break

    def predict_insight(self, text: str) -> Optional[JSONType]:
        correction_item = self.correct(text)
        corrected_text = correction_item.latest_correction
        if corrected_text != text:
            insight: JSONType = {
                "text": text,
                "corrected": corrected_text,
            }
            insight["corrections"] = correction_item.corrections
            insight["config"] = [s.get_config() for s in self.spellcheckers]
            return insight

        return None

    @staticmethod
    def is_product_valid(product: JSONType, max_errors: Optional[int] = None) -> bool:
        if max_errors is None:
            return True
        else:
            return int(product.get("unknown_ingredients_n", 0)) <= max_errors

    def correct(self, text: str) -> SpellcheckItem:
        item = SpellcheckItem(text)
        if item.is_lang_allowed:
            for spellcheck in self.spellcheckers:
                spellcheck.predict([item])
        return item



from multiprocessing import Process, SimpleQueue
from typing import Optional

import requests

from robotoff import settings
from robotoff.utils import get_logger

logger = get_logger(__name__)


class EventProcessor:
    """Send events in an outside process"""

    # the process and queue to send events
    process = None
    queue = None

    def get(self):
        """Start a process to handle events, but only when needed,
        and return communication pipe
        """
        if self.process is None:
            self.queue = SimpleQueue()
            # Create a daemonic process
            self.process = Process(target=send_events, args=(self.queue,), daemon=True)
            self.process.start()
        return self.queue

    def send_async(self, *args, **kwargs):
        if settings.EVENTS_API_URL:
            queue = self.get()
            queue.put((settings.EVENTS_API_URL, args, kwargs))


# a singleton for event processor
event_processor = EventProcessor()


def send_events(queue):
    """Loop to send events in a specific process"""
    while True:
        api_url, args, kwargs = queue.get()
        send_event(api_url, *args, **kwargs)


def send_event(
    api_url: str,
    event_type: str,
    user_id: str,
    device_id: str,
    barcode: Optional[str] = None,
):
    event = {
        "event_type": event_type,
        "user_id": user_id,
        "device_id": device_id,
        "barcode": barcode,
    }
    logger.debug("Event: %s", event)
    response = requests.post(api_url, json=event)
    logger.debug("Event API response: %s", response)
    return response



import datetime
import pathlib
from typing import Optional

from requests.exceptions import HTTPError, Timeout

from robotoff.insights.extraction import (
    DEFAULT_OCR_PREDICTION_TYPES,
    extract_ocr_predictions,
    run_object_detection_model,
)
from robotoff.insights.importer import import_insights
from robotoff.logos import (
    add_logos_to_ann,
    filter_logos,
    get_logo_confidence_thresholds,
    import_logo_insights,
    save_nearest_neighbors,
)
from robotoff.models import (
    ImageModel,
    ImagePrediction,
    LogoAnnotation,
    Prediction,
    db,
    with_db,
)
from robotoff.off import get_server_type, get_source_from_url
from robotoff.products import Product, get_product_store
from robotoff.slack import NotifierFactory
from robotoff.types import ObjectDetectionModel, PredictionType
from robotoff.utils import get_image_from_url, get_logger, http_session
from robotoff.workers.queues import enqueue_job, high_queue

logger = get_logger(__name__)


def run_import_image_job(
    barcode: str, image_url: str, ocr_url: str, server_domain: str
):
    """This job is triggered every time there is a new OCR image available for
    processing by Robotoff, via /api/v1/images/import.

    On each image import, Robotoff performs the following tasks:

    1. Generates various predictions based on the OCR-extracted text from the image.
    2. Extracts the nutriscore prediction based on the nutriscore ML model.
    3. Triggers the 'object_detection' task
    4. Stores the imported image metadata in the Robotoff DB.
    """
    logger.info(
        f"Running `import_image` for product {barcode} ({server_domain}), image {image_url}"
    )
    image = get_image_from_url(image_url, error_raise=False, session=http_session)

    if image is None:
        logger.info("Error while downloading image %s", image_url)
        return

    source_image = get_source_from_url(image_url)

    product = get_product_store()[barcode]
    if product is None:
        logger.info(
            "Product %s does not exist during image import (%s)", barcode, source_image
        )
        return

    with db:
        save_image(barcode, source_image, product, server_domain)

    enqueue_job(
        import_insights_from_image,
        high_queue,
        barcode=barcode,
        image_url=image_url,
        ocr_url=ocr_url,
        server_domain=server_domain,
    )
    # The two following tasks take longer than the previous one, so it
    # shouldn't be an issue to launch tasks concurrently (and we still have
    # the insight import lock to avoid concurrent insight import in DB for the
    # same product)
    enqueue_job(
        run_logo_object_detection,
        high_queue,
        barcode=barcode,
        image_url=image_url,
        server_domain=server_domain,
    )
    enqueue_job(
        run_nutrition_table_object_detection,
        high_queue,
        barcode=barcode,
        image_url=image_url,
        server_domain=server_domain,
    )


def import_insights_from_image(
    barcode: str,
    image_url: str,
    ocr_url: str,
    server_domain: str,
):
    image = get_image_from_url(image_url, error_raise=False, session=http_session)

    if image is None:
        logger.info("Error while downloading image %s", image_url)
        return

    source_image = get_source_from_url(image_url)
    predictions = extract_ocr_predictions(
        barcode, ocr_url, DEFAULT_OCR_PREDICTION_TYPES
    )
    if any(
        prediction.value_tag == "en:nutriscore"
        and prediction.type == PredictionType.label
        for prediction in predictions
    ):
        enqueue_job(
            run_nutriscore_object_detection,
            high_queue,
            barcode=barcode,
            image_url=image_url,
            server_domain=server_domain,
        )
    NotifierFactory.get_notifier().notify_image_flag(
        [p for p in predictions if p.type == PredictionType.image_flag],
        source_image,
        barcode,
    )

    with db:
        imported = import_insights(predictions, server_domain)
        logger.info("Import finished, %s insights imported", imported)


def save_image_job(batch: list[tuple[str, str]], server_domain: str):
    """Save a batch of images in DB.

    :param batch: a batch of (barcode, source_image) tuples
    :param server_domain: the server domain to use
    """
    with db.connection_context():
        for barcode, source_image in batch:
            product = get_product_store()[barcode]
            if product is None:
                continue

            with db.atomic():
                save_image(barcode, source_image, product, server_domain)


def save_image(
    barcode: str, source_image: str, product: Product, server_domain: str
) -> Optional[ImageModel]:
    """Save imported image details in DB."""
    if existing_image_model := ImageModel.get_or_none(source_image=source_image):
        logger.info(
            f"Image {source_image} already exist in DB, returning existing image",
        )
        return existing_image_model

    image_id = pathlib.Path(source_image).stem

    if not image_id.isdigit():
        logger.info("Non raw image was sent: %s", source_image)
        return None

    if image_id not in product.images:
        logger.info("Unknown image for product %s: %s", barcode, source_image)
        return None

    image = product.images[image_id]
    sizes = image.get("sizes", {}).get("full")

    if not sizes:
        logger.info("Image with missing size information: %s", image)
        return None

    width = sizes["w"]
    height = sizes["h"]

    if "uploaded_t" not in image:
        logger.info("Missing uploaded_t field: %s", list(image))
        return None

    uploaded_t = image["uploaded_t"]
    if isinstance(uploaded_t, str):
        if not uploaded_t.isdigit():
            logger.info("Non digit uploaded_t value: %s", uploaded_t)
            return None

        uploaded_t = int(uploaded_t)

    uploaded_at = datetime.datetime.utcfromtimestamp(uploaded_t)
    image_model = ImageModel.create(
        barcode=barcode,
        image_id=image_id,
        width=width,
        height=height,
        source_image=source_image,
        uploaded_at=uploaded_at,
        server_domain=server_domain,
        server_type=get_server_type(server_domain).name,
    )
    if image_model is not None:
        logger.info("New image %s created in DB", image_model.id)
    return image_model


def run_nutrition_table_object_detection(
    barcode: str, image_url: str, server_domain: str
):
    logger.info(
        f"Running nutrition table object detection for product {barcode} "
        f"({server_domain}), image {image_url}"
    )

    image = get_image_from_url(image_url, error_raise=False, session=http_session)

    if image is None:
        logger.info("Error while downloading image %s", image_url)
        return

    source_image = get_source_from_url(image_url)

    with db:
        run_object_detection_model(
            ObjectDetectionModel.nutrition_table, image, source_image
        )


NUTRISCORE_LABELS = {
    "nutriscore-a": "en:nutriscore-grade-a",
    "nutriscore-b": "en:nutriscore-grade-b",
    "nutriscore-c": "en:nutriscore-grade-c",
    "nutriscore-d": "en:nutriscore-grade-d",
    "nutriscore-e": "en:nutriscore-grade-e",
}


def run_nutriscore_object_detection(barcode: str, image_url: str, server_domain: str):
    logger.info(
        f"Running nutriscore object detection for product {barcode} "
        f"({server_domain}), image {image_url}"
    )

    image = get_image_from_url(image_url, error_raise=False, session=http_session)

    if image is None:
        logger.info("Error while downloading image %s", image_url)
        return

    source_image = get_source_from_url(image_url)

    with db:
        image_prediction = run_object_detection_model(
            ObjectDetectionModel.nutriscore, image, source_image
        )

    if not image_prediction:
        return

    results = [
        item for item in image_prediction.data["objects"] if item["score"] >= 0.5
    ]

    if not results:
        return

    if len(results) > 1:
        logger.info("more than one nutriscore detected, discarding detections")
        return

    result = results[0]
    score = result["score"]
    label_tag = NUTRISCORE_LABELS[result["label"]]

    with db:
        prediction = Prediction(
            type=PredictionType.label,
            barcode=barcode,
            source_image=source_image,
            value_tag=label_tag,
            automatic_processing=False,
            server_domain=server_domain,
            data={
                "confidence": score,
                "bounding_box": result["bounding_box"],
                "model": ObjectDetectionModel.nutriscore.value,
            },
        )
        import_insights([prediction], server_domain)


def run_logo_object_detection(
    barcode: str, image_url: str, server_domain: str, process_logos: bool = True
):
    """Detect logos using the universal logo detector model and generate
    logo-related predictions.

    :param barcode: Product barcode
    :param image_url: URL of the image to use
    :param server_domain: The server domain associated with the image
    :param process_logos: if True, process created logos: send them to
      Robotoff ANN and fetch nearest neighbors
    """
    logger.info(
        f"Running logo object detection for product {barcode} "
        f"({server_domain}), image {image_url}"
    )

    image = get_image_from_url(image_url, error_raise=False, session=http_session)

    if image is None:
        logger.info("Error while downloading image %s", image_url)
        return

    source_image = get_source_from_url(image_url)

    with db:
        image_prediction = run_object_detection_model(
            ObjectDetectionModel.universal_logo_detector, image, source_image
        )

        if image_prediction is None:
            # Can occur in normal conditions if an image prediction
            # already exists for this image and model
            return

        logo_ids = []
        for i, item in filter_logos(
            image_prediction.data["objects"],
            score_threshold=0.5,
            iou_threshold=0.95,
        ):
            logo_ids.append(
                LogoAnnotation.create(
                    image_prediction=image_prediction,
                    index=i,
                    score=item["score"],
                    bounding_box=item["bounding_box"],
                ).id
            )

    logger.info("%s logos found for image %s", len(logo_ids), source_image)
    if logo_ids and process_logos:
        enqueue_job(
            process_created_logos,
            high_queue,
            job_kwargs={},
            image_prediction_id=image_prediction.id,
            server_domain=server_domain,
        )


@with_db
def process_created_logos(image_prediction_id: int, server_domain: str):
    logos = (
        LogoAnnotation.select()
        .join(ImagePrediction)
        .join(ImageModel)
        .where(ImagePrediction.id == image_prediction_id)
    )

    if logos:
        image_instance = logos[0].image_prediction.image

        try:
            add_logos_to_ann(image_instance, logos)
        except (HTTPError, Timeout) as e:
            logger.info(
                "Request error during logo addition to ANN: %s, %s",
                type(e).__name__,
                e,
            )
            return

        try:
            save_nearest_neighbors(logos)
        except (HTTPError, Timeout) as e:
            logger.info(
                "Request error during ANN batch query: %s, %s",
                type(e).__name__,
                e,
            )
            return

        thresholds = get_logo_confidence_thresholds()
        import_logo_insights(logos, thresholds=thresholds, server_domain=server_domain)



import datetime
import pathlib

import tqdm

from robotoff.models import ImageModel, ImagePrediction, LogoAnnotation
from robotoff.off import generate_image_path
from robotoff.utils import get_logger, jsonl_iter

logger = get_logger(__name__)


TYPE = "object_detection"


def get_seen_set() -> set[tuple[str, str]]:
    seen_set: set[tuple[str, str]] = set()
    for prediction in (
        ImagePrediction.select(ImagePrediction.model_name, ImageModel.source_image)
        .join(ImageModel)
        .iterator()
    ):
        seen_set.add((prediction.model_name, prediction.image.source_image))

    return seen_set


def insert_batch(data_path: pathlib.Path, model_name: str, model_version: str) -> int:
    timestamp = datetime.datetime.utcnow()
    logger.info("Loading seen set...")
    seen_set = get_seen_set()
    logger.info("Seen set loaded")
    inserted = 0

    for item in tqdm.tqdm(jsonl_iter(data_path)):
        barcode = item["barcode"]
        source_image = generate_image_path(barcode=barcode, image_id=item["image_id"])
        key = (model_name, source_image)

        if key in seen_set:
            continue

        image_instance = ImageModel.get_or_none(source_image=source_image)

        if image_instance is None:
            logger.warning("Unknown image in DB: {}".format(source_image))
            continue

        results = [r for r in item["result"] if r["score"] > 0.1]
        data = {"objects": results}
        max_confidence = max([r["score"] for r in results], default=None)

        inserted += 1
        image_prediction = ImagePrediction.create(
            type=TYPE,
            image=image_instance,
            timestamp=timestamp,
            model_name=model_name,
            model_version=model_version,
            data=data,
            max_confidence=max_confidence,
        )
        for i, item in enumerate(results):
            if item["score"] >= 0.5:
                LogoAnnotation.create(
                    image_prediction=image_prediction,
                    index=i,
                    score=item["score"],
                    bounding_box=item["bounding_box"],
                )
        seen_set.add(key)

    return inserted



import abc
import datetime
import enum
import gzip
import json
import os
import pathlib
import shutil
import tempfile
from typing import Iterable, Iterator, Optional, Union

import requests
from pymongo import MongoClient

from robotoff import settings
from robotoff.mongo import MONGO_CLIENT_CACHE
from robotoff.utils import get_logger, gzip_jsonl_iter, http_session, jsonl_iter
from robotoff.utils.cache import CachedStore
from robotoff.utils.types import JSONType

logger = get_logger(__name__)


def get_image_id(image_path: str) -> Optional[str]:
    """Return the image ID from an image path.

    :param image_path: the image path (ex: /322/247/762/7888/2.jpg)
    :return: the image ID ("2" in the previous example) or None if the image
    is not "raw" (not digit-numbered)
    """
    image_id = pathlib.Path(image_path).stem

    if image_id.isdigit():
        return image_id

    return None


def is_valid_image(images: JSONType, image_path: str) -> bool:
    image_id = pathlib.Path(image_path).stem
    if not image_id.isdigit():
        return False

    return image_id in images


def is_nutrition_image(
    images: JSONType, image_path: str, lang: Optional[str] = None
) -> bool:
    return is_special_image(images, image_path, "nutrition", lang)


def has_nutrition_image(images: JSONType, lang: Optional[str] = None) -> bool:
    return has_special_image(images, "nutrition", lang)


def has_special_image(images: JSONType, key: str, lang: Optional[str] = None) -> bool:
    field_name = "{}_".format(key) if lang is None else "{}_{}".format(key, lang)
    for image_key in images:
        if image_key.startswith(field_name):
            return True

    return False


def is_special_image(
    images: JSONType, image_path: str, image_type: str, lang: Optional[str] = None
) -> bool:
    if not is_valid_image(images, image_path):
        return False

    image_id = pathlib.Path(image_path).stem

    for image_key, image_data in images.items():
        if (
            image_key.startswith(image_type)
            and str(image_data.get("imgid")) == image_id
        ):
            if lang is None:
                return True

            elif image_key.endswith("_{}".format(lang)):
                return True

    return False


def minify_product_dataset(dataset_path: pathlib.Path, output_path: pathlib.Path):
    if dataset_path.suffix == ".gz":
        jsonl_iter_func = gzip_jsonl_iter
    else:
        jsonl_iter_func = jsonl_iter

    with gzip.open(output_path, "wt", encoding="utf-8") as output_:
        for item in jsonl_iter_func(dataset_path):
            available_fields = Product.get_fields()

            minified_item = dict(
                (
                    (field, value)
                    for (field, value) in item.items()
                    if field in available_fields
                )
            )
            output_.write(json.dumps(minified_item) + "\n")


def get_product_dataset_etag() -> Optional[str]:
    if not settings.JSONL_DATASET_ETAG_PATH.is_file():
        return None

    with open(settings.JSONL_DATASET_ETAG_PATH, "r") as f:
        return f.readline()


def save_product_dataset_etag(etag: str):
    with open(settings.JSONL_DATASET_ETAG_PATH, "w") as f:
        return f.write(etag)


def fetch_dataset(minify: bool = True) -> bool:
    with tempfile.TemporaryDirectory() as tmp_dir:
        output_dir = pathlib.Path(tmp_dir)
        output_path = output_dir / "products.jsonl.gz"
        etag = download_dataset(output_path)

        logger.info("Checking dataset file integrity")
        if not is_valid_dataset(output_path):
            return False

        if minify:
            minify_path = output_dir / "products-min.jsonl.gz"
            logger.info("Minifying product JSONL")
            minify_product_dataset(output_path, minify_path)

        logger.info("Moving file(s) to dataset directory")
        shutil.copy(str(output_path), settings.JSONL_DATASET_PATH)

        if minify:
            shutil.copy(str(minify_path), settings.JSONL_MIN_DATASET_PATH)

        save_product_dataset_etag(etag)
        logger.info("Dataset fetched")
        return True


def has_dataset_changed() -> bool:
    etag = get_product_dataset_etag()

    if etag is not None:
        r = requests.head(settings.JSONL_DATASET_URL)

        current_etag = r.headers.get("ETag", "").strip("'\"")

        if current_etag == etag:
            logger.info("Dataset ETag has not changed")
            return False

    return True


def download_dataset(output_path: os.PathLike) -> str:
    r = http_session.get(settings.JSONL_DATASET_URL, stream=True)
    current_etag = r.headers.get("ETag", "").strip("'\"")

    logger.info("Dataset has changed, downloading file")
    logger.debug("Saving temporary file in {}".format(output_path))

    with open(output_path, "wb") as f:
        shutil.copyfileobj(r.raw, f)

    return current_etag


def is_valid_dataset(dataset_path: pathlib.Path) -> bool:
    """Check the dataset integrity: readable end to end and with a minimum number of products.
    This is used to spot corrupted archive files."""
    dataset = ProductDataset(dataset_path)
    try:
        count = dataset.count()
    except Exception as e:
        logger.error("Exception raised during dataset iteration", exc_info=e)
        return False

    if count < settings.DATASET_CHECK_MIN_PRODUCT_COUNT:
        logger.error(
            "Dataset has %s products, less than minimum of %s products",
            count,
            settings.DATASET_CHECK_MIN_PRODUCT_COUNT,
        )
        return False

    return True


class ComparisonOperator(enum.Enum):
    eq = 1
    gt = 2
    geq = 3
    lt = 4
    leq = 5

    @classmethod
    def get_from_string(cls, value: str):
        for operator in cls:
            if operator.name == value:
                return operator

        raise ValueError("unknown operator: {}".format(value))


def apply_comparison_operator(
    value_1, value_2, comparison_operator: ComparisonOperator
) -> bool:
    try:
        if comparison_operator == ComparisonOperator.eq:
            return value_1 == value_2

        elif comparison_operator == ComparisonOperator.gt:
            return value_1 > value_2

        elif comparison_operator == ComparisonOperator.geq:
            return value_1 >= value_2

        elif comparison_operator == ComparisonOperator.lt:
            return value_1 < value_2

        else:
            return value_1 <= value_2
    except TypeError:
        return False


class ProductStream:
    """Starting for a stream of dict representing products,
    provides a stream with filter methods to narrow down data.
    """

    def __init__(self, iterator: Iterable[JSONType]):
        self.iterator: Iterable[JSONType] = iterator

    def __iter__(self) -> Iterator[JSONType]:
        yield from self.iterator

    def filter_by_country_tag(self, country_tag: str) -> "ProductStream":
        filtered = (
            product
            for product in self.iterator
            if country_tag in (product.get("countries_tags") or [])
        )
        return ProductStream(filtered)

    def filter_by_state_tag(self, state_tag: str) -> "ProductStream":
        filtered = (
            product
            for product in self.iterator
            if state_tag in (product.get("states_tags") or [])
        )
        return ProductStream(filtered)

    def filter_text_field(self, field: str, value: str):
        filtered = (
            product for product in self.iterator if product.get(field, "") == value
        )
        return ProductStream(filtered)

    def filter_number_field(
        self,
        field: str,
        ref: Union[int, float],
        default: Union[int, float],
        operator: str = "eq",
    ) -> "ProductStream":
        operator_ = ComparisonOperator.get_from_string(operator)
        filtered = (
            product
            for product in self.iterator
            if apply_comparison_operator(product.get(field, default), ref, operator_)
        )
        return ProductStream(filtered)

    def filter_nonempty_text_field(self, field: str) -> "ProductStream":
        filtered = (
            product for product in self.iterator if (product.get(field) or "") != ""
        )
        return ProductStream(filtered)

    def filter_empty_text_field(self, field: str) -> "ProductStream":
        filtered = (
            product for product in self.iterator if not (product.get(field) or "") != ""
        )
        return ProductStream(filtered)

    def filter_nonempty_tag_field(self, field: str) -> "ProductStream":
        filtered = (product for product in self.iterator if (product.get(field) or []))
        return ProductStream(filtered)

    def filter_empty_tag_field(self, field: str) -> "ProductStream":
        filtered = (
            product for product in self.iterator if not (product.get(field) or [])
        )
        return ProductStream(filtered)

    def filter_by_modified_datetime(
        self,
        from_t: Optional[datetime.datetime] = None,
        to_t: Optional[datetime.datetime] = None,
    ):
        if from_t is None and to_t is None:
            raise ValueError("one of `from_t` or `to_t` must be provided")

        if from_t:
            from_timestamp = from_t.timestamp()
            filtered = (
                product
                for product in self.iterator
                if "last_modified_t" in product
                and product["last_modified_t"] >= from_timestamp
            )

        elif to_t:
            to_timestamp = to_t.timestamp()
            filtered = (
                product
                for product in self.iterator
                if "last_modified_t" in product
                and product["last_modified_t"] <= to_timestamp
            )

        return ProductStream(filtered)

    def take(self, count: int):
        for i, item in enumerate(self):
            if i >= count:
                break

            yield item

    def iter(self) -> Iterable[JSONType]:
        return iter(self)

    def iter_product(self) -> Iterable["Product"]:
        for item in self:
            yield Product(item)

    def collect(self) -> list[JSONType]:
        return list(self)


class ProductDataset:
    """Handles the iteration over products dataset
    contained in an eventually gziped file with one json by line.
    """

    def __init__(self, jsonl_path):
        self.jsonl_path = jsonl_path

    def stream(self) -> ProductStream:
        json_path_str = str(self.jsonl_path)

        if json_path_str.endswith(".gz"):
            iterator = gzip_jsonl_iter(json_path_str)
        else:
            iterator = jsonl_iter(json_path_str)

        return ProductStream(iterator)

    def count(self) -> int:
        count = 0
        for _ in self.stream():
            count += 1

        return count

    @classmethod
    def load(cls):
        return cls(settings.JSONL_DATASET_PATH)


class Product:
    """Product class."""

    __slots__ = (
        "barcode",
        "countries_tags",
        "categories_tags",
        "emb_codes_tags",
        "labels_tags",
        "quantity",
        "expiration_date",
        "brands_tags",
        "stores_tags",
        "unique_scans_n",
        "images",
    )

    def __init__(self, product: JSONType):
        self.barcode: Optional[str] = product.get("code")
        self.countries_tags: list[str] = product.get("countries_tags") or []
        self.categories_tags: list[str] = product.get("categories_tags") or []
        self.emb_codes_tags: list[str] = product.get("emb_codes_tags") or []
        self.labels_tags: list[str] = product.get("labels_tags") or []
        self.quantity: Optional[str] = product.get("quantity") or None
        self.expiration_date: Optional[str] = product.get("expiration_date") or None
        self.brands_tags: list[str] = product.get("brands_tags") or []
        self.stores_tags: list[str] = product.get("stores_tags") or []
        self.unique_scans_n: int = product.get("unique_scans_n") or 0
        self.images: JSONType = product.get("images") or {}

    @staticmethod
    def get_fields():
        return {
            "code",
            "countries_tags",
            "categories_tags",
            "emb_codes_tags",
            "labels_tags",
            "quantity",
            "expiration_date",
            "brands_tags",
            "stores_tags",
            "unique_scans_n",
            "images",
        }


class ProductStore(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __len__(self):
        pass

    @abc.abstractmethod
    def __getitem__(self, item):
        pass


class MemoryProductStore(ProductStore):
    def __init__(self, store: dict[str, Product]):
        self.store: dict[str, Product] = store

    def __len__(self):
        return len(self.store)

    @classmethod
    def load_from_path(cls, path: str):
        logger.info("Loading product store")
        ds = ProductDataset(path)
        stream = ds.stream()

        store: dict[str, Product] = {}

        for product in stream.iter_product():
            if product.barcode:
                store[product.barcode] = product

        return cls(store)

    @classmethod
    def load_min(cls):
        return cls.load_from_path(settings.JSONL_MIN_DATASET_PATH)

    @classmethod
    def load_full(cls):
        return cls.load_from_path(settings.JSONL_DATASET_PATH)

    def __getitem__(self, item) -> Optional[Product]:
        return self.store.get(item)

    def __iter__(self) -> Iterator[Product]:
        return iter(self.store.values())


class DBProductStore(ProductStore):
    def __init__(self, client: MongoClient):
        self.client = client
        self.db = self.client.off
        self.collection = self.db.products

    def __len__(self):
        return len(self.collection.estimated_document_count())

    def get_product(
        self, barcode: str, projection: Optional[list[str]] = None
    ) -> Optional[JSONType]:
        return self.collection.find_one({"code": barcode}, projection)

    def __getitem__(self, barcode: str) -> Optional[Product]:
        product = self.get_product(barcode)

        if product:
            return Product(product)

        return None

    def __iter__(self):
        yield from self.iter()

    def iter_product(self, projection: Optional[list[str]] = None):
        yield from (Product(p) for p in self.collection.find(projection=projection))


def load_min_dataset() -> ProductStore:
    ps = MemoryProductStore.load_min()
    logger.info("product store loaded ({} items)".format(len(ps)))
    return ps


def get_product_store() -> DBProductStore:
    mongo_client = MONGO_CLIENT_CACHE.get()
    return DBProductStore(client=mongo_client)


def get_product(
    barcode: str, projection: Optional[list[str]] = None
) -> Optional[JSONType]:
    """Get product from MongoDB.

    :param barcode: barcode of the product to fetch
    :param projection: list of fields to retrieve, if not provided all fields
    are queried
    :return: the product as a dict or None if it was not found
    """
    mongo_client = MONGO_CLIENT_CACHE.get()
    return mongo_client.off.products.find_one({"code": barcode}, projection)


CACHED_PRODUCT_STORE = CachedStore(load_min_dataset)



from enum import Enum, unique


@unique
class InsightType(str, Enum):
    """InsightType defines the type of the insight."""

    # The 'ingredient spellcheck' insight corrects the spelling in the given ingredients list.
    # NOTE: this insight currently relies on manual imports - it's possible these insights have
    # not been generated recently.
    ingredient_spellcheck = "ingredient_spellcheck"

    # The 'packager code' insight extracts the packager code using regex from the image OCR.
    # This insight is always applied automatically.
    packager_code = "packager_code"

    # The 'label' insight predicts a label that appears on the product packaging photo.
    #  Several labels are possible:
    #   - Nutriscore label detection.
    #   - Logo detection using the universal logo detector.
    #   - OCR labeling based on text parsing.
    label = "label"

    # The 'category' insight predicts the category of a product.
    # This insight can be generated by 2 different predictors:
    #  1. The Keras ML category classifier.
    #  2. The Elasticsearch category predictor based on product name.
    #
    # NOTE: there also exists a hierachical category classifier, but it hasn't generated any
    # insights since 2019.
    category = "category"

    # The 'image_flag' insight flags inappropriate images based on OCR text.
    # This insight type is never persisted to the Postgres DB and is only used to pass insight information
    # in memory.
    #
    # Currently 3 possible sources of flagged images are possible:
    #  1) "safe_search_annotation" - Google's SafeSearch API detects explicit content on product images.
    #  2) "label_annotation" - a list of hard-coded labels that should be flagged that are found on the image's OCR.
    #  3) Flashtext matches on the image's OCR.
    image_flag = "image_flag"

    # The 'product_weight' insight extracts the product weight from the image OCR.
    # This insight is never applied automatically.
    product_weight = "product_weight"

    # The 'expiration_date' insight extracts the expiration date from the image OCR.
    expiration_date = "expiration_date"

    # The 'brand' insight extracts the product's brand from the image OCR.
    brand = "brand"

    # The 'image_orientation' insight predicts the image orientation of the given image.
    image_orientation = "image_orientation"

    # The 'store' insight detects the store where the given product is sold from the image OCR.
    store = "store"

    # The 'nutrient' insight detects the list of nutrients mentioned in a product, alongside their numeric value from the image OCR.
    nutrient = "nutrient"

    # The 'trace' insight detects traces that are present in the product from the image OCR.
    # NOTE: there are 0 insights of this type in the Postgres DB.
    trace = "trace"

    # (legacy) The 'packaging_text' insight detects the type of packaging based on the image OCR.
    # This type used to be named 'packaging' and has been renamed into `packaging_text`, so that
    # 'packaging' type can be used for detections of detailed packaging information
    # (shape, material,...).
    # This insight used to be applied automatically.
    packaging_text = "packaging_text"

    # The 'packaging' insight detects the type of packaging based on the image OCR.
    # The details about each packaging element (shape, material, recycling) is returned.
    packaging = "packaging"

    # The 'location' insight detects the location of where the product comes from from the image OCR.
    # NOTE: there are 0 insights of this type in the Postgres DB.
    location = "location"

    # The 'nutrient_mention' insight detect mentions of nutrients from the image OCR (without actual values).
    nutrient_mention = "nutrient_mention"

    # The 'image_lang' insight detects which languages are mentioned on the product from the image OCR.
    image_lang = "image_lang"

    # The 'nutrition_image' insight tags images that have nutrition information based on the 'nutrient_mention' insight and the 'image_orientation' insight.
    # NOTE: this insight and the dependant insights has not been generated since 2020.
    nutrition_image = "nutrition_image"

    # The 'nutritional_table_structure' insight detects the nutritional table structure from the image.
    # NOTE: this insight has not been generated since 2020.
    nutrition_table_structure = "nutrition_table_structure"



import re

from robotoff.spellcheck.base_spellchecker import BaseSpellchecker

PERCENTAGE_REGEX = re.compile(
    r"(\A|.)([0-9]{0,2})([ ]{0,1}?[,|.|;|/]{0,1}[ ]{0,1})([0-9]{0,2})[ ]?(?:%|(?:[\?|/|\\](?:\D|\Z)))"
)
ADDITIVES_REGEX = re.compile(r"(?:E ?\d{3,5}[a-z]*)", re.IGNORECASE)
VERSION = "1"


class PercentagesSpellchecker(BaseSpellchecker):
    def correct(self, text: str) -> str:
        return self.format_percentages(text)

    @staticmethod
    def format_percentages(text: str) -> str:
        """
        Rule-based formatting of the percentages in a string.
        Related to this repository : https://github.com/openfoodfacts/openfoodfacts-ai/tree/master/spellcheck
        In particular, please read Readme + tests implemented.

        Rules applied :
            - Match substrings following this pattern :
                [0 to 2 digits][(optionally) a whitespace][a separator][0 to 2 digits][(optionally) a whitespace][a % or similar symbol].
            - If the match contains/is part of an additive (e.g. Exxx), we drop it and to nothing.
            - Depending on whether the first or last digits are present or not, we format the match accordingly (e.g XX,XX% or XX%).
            - If no separator is found (only a whitespace), we concatenate the digits (example : 4 0% -> 40%). Just as a guarantee, we make sure that the value is below (or equals) to 100.
            - In addition to these rules, we pad the match with whitespaces if context needs it. Pad is added if previous or next char is an alphanumerical character (raisin7% -> raisin 7%) or an opening/closing parenthesis (19%(lait -> 19% (lait).

        Some examples :
            - "AOP (lait) 3 ,5%, sirop de" -> "AOP (lait) 3,5%, sirop de"
            - "100 % Coco sucre"           -> "100% Coco sucre"
            - "fumée 17.1% [viande de"     -> "fumée 17,1% [viande de"
            - "concentré 13 %, acidifiant" -> "concentré 13%, acidifiant"
        """
        formatted_text_list: list[str] = []
        last_index: int = 0
        for match in PERCENTAGE_REGEX.finditer(text):
            first_char, first_digits, sep, last_digits = match.groups()

            start = match.start() + len(first_char)
            end = match.end()  # - len(to_drop)
            nb_first_digits = len(first_digits)
            nb_last_digits = len(last_digits)

            pad_before = False
            pad_after = False

            if ADDITIVES_REGEX.match(text[match.start() : match.end()]):
                # Very conservative rule
                formatted_match = text[start:end]

            elif nb_first_digits == 0 and nb_last_digits == 0:
                # Not a good match
                formatted_match = text[start:end]

            elif nb_first_digits == 0:
                formatted_match = f"{sep}{last_digits}%"
                pad_before = False
                pad_after = True

            elif nb_last_digits == 0:
                formatted_match = f"{first_digits}%"
                pad_before = True
                pad_after = True

            elif len(sep) > 0 and (nb_first_digits == 2 or nb_last_digits == 2):
                formatted_match = f"{first_digits},{last_digits}%"
                pad_before = True
                pad_after = True

            elif sep.strip() == "":
                if float(f"{first_digits}{last_digits}") <= 100.0:
                    formatted_match = f"{first_digits}{last_digits}%"
                    pad_before = True
                    pad_after = True
                else:
                    formatted_match = f"{first_digits},{last_digits}%"
                    pad_before = True
                    pad_after = True
            else:
                formatted_match = f"{first_digits},{last_digits}%"
                pad_before = True
                pad_after = True

            if pad_before:
                if start > 0:
                    previous_char = text[start - 1]
                    if previous_char.isalnum() or previous_char in ["*", ")", "]", "}"]:
                        formatted_match = " " + formatted_match.lstrip(" ")

            if pad_after:
                if end < len(text):
                    next_char = text[end]
                    if next_char.isalnum() or next_char in ["*", "(", "[", "{"]:
                        formatted_match = formatted_match.rstrip(" ") + " "

            formatted_text_list.append(text[last_index:start])
            formatted_text_list.append(formatted_match)
            last_index = end
        formatted_text_list.append(text[last_index:])

        return "".join(formatted_text_list)

    def get_config(self):
        return {
            "version": VERSION,
            "name": self.__class__.__name__,
        }



import datetime
from typing import Callable, Optional

from robotoff.utils import get_logger

logger = get_logger(__name__)


class CachedStore:
    def __init__(self, fetch_func: Callable, expiration_interval: Optional[int] = 30):
        self.store = None
        self.expires_after: Optional[datetime.datetime] = None
        self.fetch_func: Callable = fetch_func
        self.expiration_timedelta: Optional[datetime.timedelta]

        if expiration_interval is not None:
            self.expiration_timedelta = datetime.timedelta(minutes=expiration_interval)
        else:
            self.expiration_timedelta = None

    def get(self, **kwargs):
        if self.store is None or (
            self.expiration_timedelta is not None
            and datetime.datetime.utcnow() >= self.expires_after
        ):
            if self.store is not None:
                logger.info("CachedStore expired, reloading...")

            if self.expiration_timedelta is not None:
                self.expires_after = (
                    datetime.datetime.utcnow() + self.expiration_timedelta
                )
            self.store = self.fetch_func(**kwargs)

        return self.store



import operator
from typing import Optional

import cachetools
import numpy as np

from robotoff import settings
from robotoff.insights.importer import import_insights
from robotoff.logo_label_type import LogoLabelType
from robotoff.models import ImageModel, LogoAnnotation, LogoConfidenceThreshold
from robotoff.prediction.types import Prediction
from robotoff.slack import NotifierFactory
from robotoff.types import PredictionType
from robotoff.utils import get_logger, http_session
from robotoff.utils.types import JSONType

logger = get_logger(__name__)


LOGO_TYPE_MAPPING: dict[str, PredictionType] = {
    "brand": PredictionType.brand,
    "label": PredictionType.label,
}

UNKNOWN_LABEL: LogoLabelType = ("UNKNOWN", None)


BoundingBoxType = tuple[float, float, float, float]


def load_resources():
    """Load and cache resources."""
    logger.info("Loading logo resources...")
    get_logo_confidence_thresholds()
    get_logo_annotations()


def compute_iou(box_1: BoundingBoxType, box_2: BoundingBoxType) -> float:
    """Compute the IoU (intersection over union) for two bounding boxes.

    The boxes are expected to have the following format:
    (y_min, x_min, y_max, x_max).
    """
    y_min_1, x_min_1, y_max_1, x_max_1 = box_1
    y_min_2, x_min_2, y_max_2, x_max_2 = box_2
    x_max = min(x_max_1, x_max_2)
    x_min = max(x_min_1, x_min_2)
    y_max = min(y_max_1, y_max_2)
    y_min = max(y_min_1, y_min_2)
    width_inter = max(0, x_max - x_min)
    height_inter = max(0, y_max - y_min)
    area_inter = width_inter * height_inter
    box_1_area = (x_max_1 - x_min_1) * (y_max_1 - y_min_1)
    box_2_area = (x_max_2 - x_min_2) * (y_max_2 - y_min_2)
    union_area = box_1_area + box_2_area - area_inter
    return area_inter / union_area


def filter_logos(
    logos: list[JSONType], score_threshold: float, iou_threshold: float = 0.95
) -> list[tuple[int, JSONType]]:
    """Select logos that don't intersect with each other
    (IoU < `iou_threshold`) and that have a confidence score above
    `score_threshold`.

    Return a list of (original_idx, logo) tuples.
    """
    filtered = []
    skip_indexes = set()
    for i in range(len(logos)):
        logo = logos[i]
        if i not in skip_indexes:
            for j in range(i + 1, len(logos)):
                if (
                    compute_iou(logo["bounding_box"], logos[j]["bounding_box"])
                    >= iou_threshold
                ):
                    # logos are sorted by descending confidence score, so we ignore
                    # j logo (logo with lower confidence score)
                    skip_indexes.add(j)

        if logo["score"] >= score_threshold:
            filtered.append((i, logo))

    return filtered


@cachetools.cached(cachetools.LRUCache(maxsize=1))
def get_logo_confidence_thresholds() -> dict[LogoLabelType, float]:
    thresholds = {}

    for item in LogoConfidenceThreshold.select().iterator():
        thresholds[(item.type, item.value)] = item.threshold

    return thresholds


def get_stored_logo_ids() -> set[int]:
    r = http_session.get(
        settings.BaseURLProvider().robotoff().get() + "/api/v1/ann/stored", timeout=30
    )

    if not r.ok:
        logger.warning(
            f"error while fetching stored logo IDs ({r.status_code}): %s", r.text
        )
        return set()

    return set(r.json()["stored"])


def add_logos_to_ann(image: ImageModel, logos: list[LogoAnnotation]) -> int:
    if not logos:
        return 0

    image_url = settings.OFF_IMAGE_BASE_URL + image.source_image

    data = {
        "image_url": image_url,
        "logos": [{"bounding_box": logo.bounding_box, "id": logo.id} for logo in logos],
    }
    r = http_session.post(
        settings.BaseURLProvider().robotoff().get() + "/api/v1/ann/add",
        json=data,
        timeout=30,
    )

    r.raise_for_status()
    return r.json()["added"]


def save_nearest_neighbors(logos: list[LogoAnnotation]) -> int:
    logo_ids_params = ",".join((str(logo.id) for logo in logos))
    r = http_session.get(
        settings.BaseURLProvider().robotoff().get()
        + "/api/v1/ann/batch?logo_ids="
        + logo_ids_params,
        timeout=30,
    )

    r.raise_for_status()
    response = r.json()

    results = {int(key): value for key, value in response["results"].items()}

    logo_id_to_logo = {logo.id: logo for logo in logos}
    missing_logo_ids = set(logo_id_to_logo.keys()).difference(set(results.keys()))

    if missing_logo_ids:
        logger.info("Missing logo IDs in response: %s", missing_logo_ids)

    saved = 0
    for logo_id, logo_results in results.items():
        if logo_id in logo_id_to_logo:
            logo = logo_id_to_logo[logo_id]
            distances = [n["distance"] for n in logo_results]
            logo_ids = [n["logo_id"] for n in logo_results]
            logo.nearest_neighbors = {
                "distances": distances,
                "logo_ids": logo_ids,
            }
            logo.save()
            saved += 1

    return saved


@cachetools.cached(cachetools.LRUCache(maxsize=1))
def get_logo_annotations() -> dict[int, LogoLabelType]:
    annotations: dict[int, LogoLabelType] = {}

    for logo in (
        LogoAnnotation.select(
            LogoAnnotation.id,
            LogoAnnotation.annotation_type,
            LogoAnnotation.annotation_value,
            LogoAnnotation.taxonomy_value,
        )
        .where(LogoAnnotation.annotation_type.is_null(False))
        .iterator()
    ):
        if logo.annotation_value is None:
            annotations[logo.id] = (logo.annotation_type, None)
        elif logo.taxonomy_value is not None:
            annotations[logo.id] = (logo.annotation_type, logo.taxonomy_value)

    return annotations


def predict_label(logo: LogoAnnotation) -> Optional[LogoLabelType]:
    probs = predict_proba(logo)

    if probs is None or not probs:
        return None

    return sorted(probs.items(), key=operator.itemgetter(0))[0][0]


def predict_proba(
    logo: LogoAnnotation, weights: str = "distance"
) -> Optional[dict[LogoLabelType, float]]:
    if logo.nearest_neighbors is None:
        return None

    nn_distances = logo.nearest_neighbors["distances"]
    nn_logo_ids = logo.nearest_neighbors["logo_ids"]

    logo_annotations = get_logo_annotations()

    nn_labels: list[LogoLabelType] = []
    for nn_logo_id in nn_logo_ids:
        nn_labels.append(logo_annotations.get(nn_logo_id, UNKNOWN_LABEL))

    return _predict_proba(nn_logo_ids, nn_labels, nn_distances, weights)


def _predict_proba(
    logo_ids: list[int],
    nn_labels: list[LogoLabelType],
    nn_distances: list[float],
    weights: str,
) -> dict[LogoLabelType, float]:
    weights = get_weights(np.array(nn_distances), weights)
    labels: list[LogoLabelType] = [UNKNOWN_LABEL] + [
        x for x in set(nn_labels) if x != UNKNOWN_LABEL
    ]
    label_to_id = {label: i for i, label in enumerate(labels)}
    proba_k = np.zeros(len(labels))
    pred_labels = np.array([label_to_id[x] for x in nn_labels])

    for i, idx in enumerate(pred_labels.T):
        proba_k[idx] += weights[i]

    proba_k /= proba_k.sum()

    prediction: dict[LogoLabelType, float] = {}
    for i in range(len(proba_k)):
        prediction[labels[i]] = float(proba_k[i])

    return prediction


def get_weights(dist: np.ndarray, weights: str = "uniform"):
    """Get the weights from an array of distances and a parameter ``weights``
    Parameters
    ----------
    dist : ndarray
        The input distances
    weights : {'uniform', 'distance' or a callable}
        The kind of weighting used
    Returns
    -------
    weights_arr : array of the same shape as ``dist``
    """
    if weights in (None, "uniform"):
        return np.ones_like(dist)

    elif weights == "distance":
        # if user attempts to classify a point that was zero distance from one
        # or more training points, those training points are weighted as 1.0
        # and the other points as 0.0
        with np.errstate(divide="ignore"):
            dist = 1.0 / dist
        inf_mask = np.isinf(dist)
        inf_row = np.any(inf_mask)
        dist[inf_row] = inf_mask[inf_row]
        return dist
    elif callable(weights):
        return weights(dist)
    else:
        raise ValueError(
            "weights not recognized: should be 'uniform', "
            "'distance', or a callable function"
        )


def import_logo_insights(
    logos: list[LogoAnnotation],
    server_domain: str,
    thresholds: dict[LogoLabelType, float],
    default_threshold: float = 0.1,
):
    selected_logos = []
    logo_probs = []
    for logo in logos:
        probs = predict_proba(logo)

        if not probs:
            continue

        label, max_prob = max(
            ((label, prob) for label, prob in probs.items() if label != UNKNOWN_LABEL),
            default=(UNKNOWN_LABEL, 0.0),
            key=operator.itemgetter(1),
        )
        threshold = thresholds[label] if label in thresholds else default_threshold

        if label == UNKNOWN_LABEL or max_prob < threshold:
            continue

        selected_logos.append(logo)
        logo_probs.append(probs)

    predictions = predict_logo_predictions(selected_logos, logo_probs)
    imported = import_insights(predictions, server_domain)

    for logo, probs in zip(selected_logos, logo_probs):
        NotifierFactory.get_notifier().send_logo_notification(logo, probs)

    return imported


def generate_insights_from_annotated_logos(
    logos: list[LogoAnnotation],
    server_domain: str,
) -> int:
    predictions = []
    for logo in logos:
        prediction = generate_prediction(
            logo_type=logo.annotation_type,
            logo_value=logo.taxonomy_value,
            automatic_processing=True,  # because this is a user annotation, which we trust.
            data={
                "confidence": 1.0,
                "logo_id": logo.id,
                "bounding_box": logo.bounding_box,
                "username": logo.username,
                "is_annotation": True,  # it's worth restating it
            },
        )

        if prediction is None:
            continue

        image = logo.image_prediction.image
        prediction.barcode = image.barcode
        prediction.source_image = image.source_image
        predictions.append(prediction)

    imported = import_insights(predictions, server_domain)

    if imported:
        logger.info("%s logo insights imported after annotation", imported)
    return imported


def predict_logo_predictions(
    logos: list[LogoAnnotation],
    logo_probs: list[dict[LogoLabelType, float]],
) -> list[Prediction]:
    predictions = []

    for logo, probs in zip(logos, logo_probs):
        if not probs:
            continue

        label, max_prob = max(
            ((label, prob) for label, prob in probs.items() if label != UNKNOWN_LABEL),
            default=(UNKNOWN_LABEL, 0.0),
            key=operator.itemgetter(1),
        )

        if label == UNKNOWN_LABEL:
            continue

        prediction = generate_prediction(
            logo_type=label[0],
            logo_value=label[1],
            data={
                "confidence": max_prob,
                "logo_id": logo.id,
                "bounding_box": logo.bounding_box,
            },
        )

        if prediction is not None:
            image = logo.image_prediction.image
            prediction.barcode = image.barcode
            prediction.source_image = image.source_image
            predictions.append(prediction)

    return predictions


def generate_prediction(
    logo_type: str,
    logo_value: Optional[str],
    data: dict,
    automatic_processing: Optional[bool] = False,
) -> Optional[Prediction]:
    """Generate a Prediction from a logo.

    The Prediction may either be created after the annotation of the logo by
    a human (in which case `automatic_processing` is True), or by infering the
    logo value from nearest neighbor labels (in which case
    `automatic_processing` is False).

    Currently, only brand and label logo types are supported: None is returned
    if the logo type is different, or if the logo_value is None.
    """
    if logo_type not in LOGO_TYPE_MAPPING or logo_value is None:
        return None

    prediction_type = LOGO_TYPE_MAPPING[logo_type]

    value_tag = None
    value = None

    if prediction_type == PredictionType.brand:
        value_tag = value = logo_value

    elif prediction_type == PredictionType.label:
        value_tag = logo_value

    return Prediction(
        type=prediction_type,
        value_tag=value_tag,
        value=value,
        automatic_processing=automatic_processing,
        predictor="universal-logo-detector",
        data=data,
    )



import gzip
import logging
import os
import pathlib
import sys
from io import BytesIO
from typing import Any, Callable, Iterable, Optional, Union

import orjson
import PIL
import requests
from PIL import Image
from requests.adapters import HTTPAdapter
from requests.exceptions import SSLError, Timeout

from robotoff import settings


def get_logger(name=None, level: Optional[int] = None):
    logger = logging.getLogger(name)

    if level is None:
        log_level = os.environ.get("LOG_LEVEL", "INFO").upper()
        level = logging.getLevelName(log_level)

        if not isinstance(level, int):
            print(
                "Unknown log level: {}, fallback to INFO".format(log_level),
                file=sys.stderr,
            )
            level = 20

    logger.setLevel(level)

    if name is None:
        configure_root_logger(logger, level)

    return logger


def configure_root_logger(logger, level: int = 20):
    logger.setLevel(level)
    handler = logging.StreamHandler()
    formatter = logging.Formatter(
        "%(asctime)s :: %(processName)s :: "
        "%(threadName)s :: %(levelname)s :: "
        "%(message)s"
    )
    handler.setFormatter(formatter)
    handler.setLevel(level)
    logger.addHandler(handler)

    for name in ("redis_lock", "spacy"):
        logging.getLogger(name).setLevel(logging.WARNING)


logger = get_logger(__name__)


def jsonl_iter(jsonl_path: Union[str, pathlib.Path]) -> Iterable[dict]:
    open_fn = get_open_fn(jsonl_path)

    with open_fn(str(jsonl_path), "rt", encoding="utf-8") as f:
        yield from jsonl_iter_fp(f)


def gzip_jsonl_iter(jsonl_path: Union[str, pathlib.Path]) -> Iterable[dict]:
    with gzip.open(jsonl_path, "rt", encoding="utf-8") as f:
        yield from jsonl_iter_fp(f)


def jsonl_iter_fp(fp) -> Iterable[dict]:
    for line in fp:
        line = line.strip("\n")
        if line:
            yield orjson.loads(line)


def load_json(
    path: Union[str, pathlib.Path], compressed: bool = False
) -> Union[dict, list]:
    """Load a JSON file.

    :param path: the path of the file
    :param: compressed: if True, use gzip to decompress the file
    :return: the unserialized JSON
    """
    if compressed:
        with gzip.open(str(path), "rb") as f:
            return orjson.loads(f.read())
    else:
        with open(str(path), "rb") as f:
            return orjson.loads(f.read())


def dump_json(path: Union[str, pathlib.Path], item: Any, compressed: bool = False):
    """Dump an object in a JSON file.

    :param path: the path of the file
    :param item: the item to serialize
    :param: compressed: if True, use gzip to compress the file
    """
    if compressed:
        with gzip.open(str(path), "wb") as f:
            f.write(orjson.dumps(item))
    else:
        with open(str(path), "wb") as f:
            f.write(orjson.dumps(item))


def dump_jsonl(
    filepath: Union[str, pathlib.Path],
    json_iter: Iterable[Any],
) -> int:
    count = 0
    open_fn = get_open_fn(filepath)

    with open_fn(str(filepath), "wb") as f:
        for item in json_iter:
            f.write(orjson.dumps(item) + b"\n")
            count += 1

    return count


def get_open_fn(filepath: Union[str, pathlib.Path]) -> Callable:
    filepath = str(filepath)
    if filepath.endswith(".gz"):
        return gzip.open
    else:
        return open


def text_file_iter(
    filepath: Union[str, pathlib.Path], comment: bool = True
) -> Iterable[str]:
    open_fn = get_open_fn(filepath)

    with open_fn(str(filepath), "rt") as f:
        for item in f:
            item = item.strip("\n")

            if item:
                # commented lines start with '//'
                if not comment or not item.startswith("//"):
                    yield item


def dump_text(filepath: Union[str, pathlib.Path], text_iter: Iterable[str]):
    open_fn = get_open_fn(filepath)

    with open_fn(str(filepath), "wt") as f:
        for item in text_iter:
            item = item.strip("\n")
            f.write(item + "\n")


class ImageLoadingException(Exception):
    """Exception raised by `get_image_from_url`` when image cannot be fetched
    from URL or if loading failed.
    """

    pass


def get_image_from_url(
    image_url: str,
    error_raise: bool = True,
    session: Optional[requests.Session] = None,
) -> Optional[Image.Image]:
    """Fetch an image from `image_url` and load it.

    :param image_url: URL of the image to load
    :param error_raise: if True, raises a `ImageLoadingException` if an error
    occured, defaults to False. If False, None is returned if an error occurs.
    :param session: requests Session to use, by default no session is used.
    :raises ImageLoadingException: _description_
    :return: the Pillow Image or None.
    """
    try:
        if session:
            r = session.get(image_url)
        else:
            r = requests.get(image_url)
    except (ConnectionError, SSLError, Timeout) as e:
        error_message = f"Cannot download image {image_url}: {type(e).__name__}, {e}"
        if error_raise:
            raise ImageLoadingException(error_message) from e
        logger.info(error_message)
        return None

    if not r.ok:
        error_message = f"Cannot download image {image_url}: HTTP {r.status_code}"
        if error_raise:
            raise ImageLoadingException(error_message)
        logger.info(error_message)
        return None

    try:
        return Image.open(BytesIO(r.content))
    except PIL.UnidentifiedImageError:
        error_message = f"Cannot identify image {image_url}"
        if error_raise:
            raise ImageLoadingException(error_message)
        logger.info(error_message)
    except PIL.Image.DecompressionBombError:
        error_message = f"Decompression bomb error for image {image_url}"
        if error_raise:
            raise ImageLoadingException(error_message)
        logger.info(error_message)

    return None


http_session = requests.Session()
USER_AGENT_HEADERS = {
    "User-Agent": settings.ROBOTOFF_USER_AGENT,
}
http_session.headers.update(USER_AGENT_HEADERS)
static_adapter = HTTPAdapter(max_retries=3)
http_session.mount("https://static.openfoodfacts.", static_adapter)
http_session.mount("https://images.openfoodfacts.", static_adapter)



from typing import Optional

from redis import Redis
from redis_lock import Lock as BaseLock

from robotoff import settings

redis_conn = Redis(host=settings.REDIS_HOST)


class LockedResourceException(Exception):
    pass


class Lock(BaseLock):
    _enabled = True

    def __init__(
        self,
        name: str,
        blocking: bool = False,
        timeout: Optional[float] = None,
        expire: int = 60,
        **kwargs,
    ):
        self.timeout = timeout
        if timeout is not None:
            blocking = True
        self.blocking = blocking
        if self._enabled:
            super().__init__(redis_conn, name=name, expire=expire, **kwargs)

    def __enter__(self):
        if self._enabled:
            acquired = self.acquire(blocking=self.blocking, timeout=self.timeout)
            if not acquired:
                raise LockedResourceException()
            return self

    def __exit__(self, *args, **kwargs):
        if self._enabled:
            self.release()



import dataclasses
import pathlib
from typing import Optional

import cachetools
import grpc
import numpy as np
from PIL import Image
from tritonclient.grpc import service_pb2, service_pb2_grpc

from robotoff import settings
from robotoff.prediction.object_detection.utils import visualization_utils as vis_util
from robotoff.types import ObjectDetectionModel
from robotoff.utils import get_logger, text_file_iter
from robotoff.utils.types import JSONType

logger = get_logger(__name__)

LABEL_NAMES_FILENAME = "labels.txt"


OBJECT_DETECTION_MODEL_VERSION = {
    ObjectDetectionModel.nutriscore: "tf-nutriscore-1.0",
    ObjectDetectionModel.nutrition_table: "tf-nutrition-table-1.0",
    ObjectDetectionModel.universal_logo_detector: "tf-universal-logo-detector-1.0",
}


@dataclasses.dataclass
class ObjectDetectionResult:
    bounding_box: tuple
    score: float
    label: str


@dataclasses.dataclass
class ObjectDetectionRawResult:
    num_detections: int
    detection_boxes: np.ndarray
    detection_scores: np.ndarray
    detection_classes: np.ndarray
    label_names: list[str]
    detection_masks: Optional[np.ndarray] = None
    boxed_image: Optional[Image.Image] = None

    def select(self, threshold: Optional[float] = None) -> list[ObjectDetectionResult]:
        if threshold is None:
            threshold = 0.5

        box_masks = self.detection_scores > threshold
        selected_boxes = self.detection_boxes[box_masks]
        selected_scores = self.detection_scores[box_masks]
        selected_classes = self.detection_classes[box_masks]

        results = []
        for bounding_box, score, label in zip(
            selected_boxes, selected_scores, selected_classes
        ):
            label_int = int(label)
            label_str = self.label_names[label_int]
            if label_str is not None:
                result = ObjectDetectionResult(
                    bounding_box=tuple(bounding_box.tolist()),
                    score=float(score),
                    label=label_str,
                )
                results.append(result)

        return results

    def to_json(self, threshold: Optional[float] = None) -> list[JSONType]:
        return [dataclasses.asdict(r) for r in self.select(threshold)]


def convert_image_to_array(image: Image.Image) -> np.ndarray:
    if image.mode != "RGB":
        image = image.convert("RGB")

    (im_width, im_height) = image.size

    return np.array(image.getdata()).reshape((im_height, im_width, 3)).astype(np.uint8)


def add_boxes_and_labels(image_array: np.ndarray, raw_result: ObjectDetectionRawResult):
    vis_util.visualize_boxes_and_labels_on_image_array(
        image_array,
        raw_result.detection_boxes,
        raw_result.detection_classes,
        raw_result.detection_scores,
        raw_result.label_names,
        instance_masks=raw_result.detection_masks,
        use_normalized_coordinates=True,
        line_thickness=5,
    )
    image_with_boxes = Image.fromarray(image_array)
    raw_result.boxed_image = image_with_boxes


def resize_image(image: Image.Image, max_size: tuple[int, int]) -> Image.Image:
    width, height = image.size
    max_width, max_height = max_size

    if width > max_width or height > max_height:
        new_image = image.copy()
        new_image.thumbnail((max_width, max_height))
        return new_image

    return image


@cachetools.cached(cachetools.Cache(maxsize=1))
def get_triton_inference_stub():
    channel = grpc.insecure_channel(settings.TRITON_URI)
    return service_pb2_grpc.GRPCInferenceServiceStub(channel)


class RemoteModel:
    def __init__(self, name: str, label_names: list[str]):
        self.name: str = name
        self.label_names = label_names

    def detect_from_image(
        self, image: Image.Image, output_image: bool = False
    ) -> ObjectDetectionRawResult:
        resized_image = resize_image(image, settings.OBJECT_DETECTION_IMAGE_MAX_SIZE)
        image_array = convert_image_to_array(resized_image)
        grpc_stub = get_triton_inference_stub()
        request = service_pb2.ModelInferRequest()
        request.model_name = self.name

        image_input = service_pb2.ModelInferRequest().InferInputTensor()
        image_input.name = "inputs"
        image_input.datatype = "UINT8"
        image_input.shape.extend([1, image_array.shape[0], image_array.shape[1], 3])
        request.inputs.extend([image_input])

        for output_name in (
            "num_detections",
            "detection_classes",
            "detection_scores",
            "detection_boxes",
        ):
            output = service_pb2.ModelInferRequest().InferRequestedOutputTensor()
            output.name = output_name
            request.outputs.extend([output])

        request.raw_input_contents.extend([image_array.tobytes()])
        response = grpc_stub.ModelInfer(request)

        if len(response.outputs) != 4:
            raise Exception(f"expected 4 output, got {len(response.outputs)}")

        if len(response.raw_output_contents) != 4:
            raise Exception(
                f"expected 4 raw output content, got {len(response.raw_output_contents)}"
            )

        output_index = {output.name: i for i, output in enumerate(response.outputs)}
        num_detections = (
            np.frombuffer(
                response.raw_output_contents[output_index["num_detections"]],
                dtype=np.float32,
            )
            .reshape((1, 1))
            .astype(np.int)[0][0]  # type: ignore
        )
        detection_scores = np.frombuffer(
            response.raw_output_contents[output_index["detection_scores"]],
            dtype=np.float32,
        ).reshape((1, -1))[0]
        detection_classes = (
            np.frombuffer(
                response.raw_output_contents[output_index["detection_classes"]],
                dtype=np.float32,
            )
            .reshape((1, -1))
            .astype(np.int)  # type: ignore
        )[0]
        detection_boxes = np.frombuffer(
            response.raw_output_contents[output_index["detection_boxes"]],
            dtype=np.float32,
        ).reshape((1, -1, 4))[0]

        result = ObjectDetectionRawResult(
            num_detections=num_detections,
            detection_classes=detection_classes,
            detection_boxes=detection_boxes,
            detection_scores=detection_scores,
            detection_masks=None,
            label_names=self.label_names,
        )

        if output_image:
            add_boxes_and_labels(image_array, result)

        return result


class ObjectDetectionModelRegistry:
    models: dict[str, RemoteModel] = {}
    _loaded = False

    @classmethod
    def get_available_models(cls) -> list[str]:
        cls.load_all()
        return list(cls.models.keys())

    @classmethod
    def load_all(cls):
        if cls._loaded:
            return
        for model in ObjectDetectionModel:
            model_name = model.value
            file_path = settings.MODELS_DIR / model_name
            if file_path.is_dir():
                logger.info("Model %s found", model_name)
                cls.models[model_name] = cls.load(model_name, file_path)
            else:
                logger.info("Missing model: %s", model_name)
        cls._loaded = True

    @classmethod
    def load(cls, name: str, model_dir: pathlib.Path) -> RemoteModel:
        label_names = ["NULL"] + list(text_file_iter(model_dir / LABEL_NAMES_FILENAME))
        model = RemoteModel(name, label_names)
        cls.models[name] = model
        return model

    @classmethod
    def get(cls, name: str) -> RemoteModel:
        cls.load_all()
        return cls.models[name]



import re
from typing import Optional, Union

from flashtext import KeywordProcessor

from robotoff import settings
from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType
from robotoff.utils import text_file_iter
from robotoff.utils.cache import CachedStore

from .dataclass import OCRField, OCRRegex, OCRResult, get_text
from .utils import generate_keyword_processor


def process_fr_packaging_match(match) -> str:
    approval_numbers = match.group(1, 2, 3)
    return "FR {}.{}.{} EC".format(*approval_numbers).upper()


def process_de_packaging_match(match) -> str:
    federal_state_tag, company_tag = match.group(1, 2)
    return "DE {}-{} EC".format(federal_state_tag, company_tag).upper()


def process_fr_emb_match(match) -> str:
    city_code, company_code = match.group(1, 2)
    city_code = city_code.replace(" ", "")
    company_code = company_code or ""
    return "EMB {}{}".format(city_code, company_code).upper()


def process_fsc_match(match) -> str:
    fsc_code = match.group(1)
    return "FSC-{}".format(fsc_code).upper()


def process_USDA_match_to_flashtext(match) -> Optional[str]:
    """Returns the USDA code matched by REGEX the same way it exists
    in the USDA database (1st column of USDA_code_flashtext.txt)
    """

    unchecked_code = match.group().upper()
    unchecked_code = re.sub(r"\s*\.*", "", unchecked_code)

    processor = USDA_CODE_KEYWORD_PROCESSOR_STORE.get()
    USDA_code = extract_USDA_code(processor, unchecked_code)
    return USDA_code


def generate_USDA_code_keyword_processor() -> KeywordProcessor:
    """Builds the KeyWordProcessor for USDA codes

    This will be called only once thanks to CachedStore
    """

    codes = text_file_iter(settings.OCR_USDA_CODE_FLASHTEXT_DATA_PATH)
    return generate_keyword_processor(codes)


def extract_USDA_code(processor: KeywordProcessor, text: str) -> Optional[str]:
    """Given a string, returns the USDA code it contains or None"""
    USDA_code = None

    for (USDA_code_keyword, _) in processor.extract_keywords(text):
        USDA_code = USDA_code_keyword
        # Once we found a match we can return the code
        # as there should not be more than one match
        break
    return USDA_code


USDA_CODE_KEYWORD_PROCESSOR_STORE = CachedStore(
    fetch_func=generate_USDA_code_keyword_processor, expiration_interval=None
)


PACKAGER_CODE = {
    "fr_emb": [
        OCRRegex(
            re.compile(r"emb ?(\d ?\d ?\d ?\d ?\d) ?([a-z])?(?![a-z0-9])"),
            field=OCRField.text_annotations,
            lowercase=True,
            processing_func=process_fr_emb_match,
        ),
    ],
    "fsc": [
        OCRRegex(
            re.compile(r"fsc.? ?(c\d{6})"),
            field=OCRField.text_annotations,
            lowercase=True,
            processing_func=process_fsc_match,
        ),
    ],
    "eu_fr": [
        OCRRegex(
            re.compile(
                r"fr (\d{2,3}|2[ab])[\-\s.](\d{3})[\-\s.](\d{3}) (ce|ec)(?![a-z0-9])"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
            processing_func=process_fr_packaging_match,
        ),
    ],
    "eu_de": [
        OCRRegex(
            re.compile(
                r"de (bb|be|bw|by|hb|he|hh|mv|ni|nw|rp|sh|sl|sn|st|th)[\-\s.](\d{1,5})[\-\s.] ?(eg|ec)(?![a-z0-9])"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
            processing_func=process_de_packaging_match,
        ),
    ],
    "rspo": [
        OCRRegex(
            re.compile(r"(?<!\w)RSPO-\d{7}(?!\d)"),
            field=OCRField.full_text_contiguous,
            lowercase=False,
        ),
    ],
    "fr_gluten": [
        OCRRegex(
            re.compile(r"FR-\d{3}-\d{3}"),
            field=OCRField.full_text_contiguous,
            lowercase=False,
        ),
    ],
    # Temporarily disable USDA extraction until the overmatching bug is fixed
    # "USDA": [
    #     # To match the USDA like "EST. 522"
    #     OCRRegex(
    #         re.compile(r"EST\.*\s*\d{1,5}[A-Z]{0,3}\.*"),
    #         field=OCRField.full_text_contiguous,
    #         lowercase=False,
    #         processing_func=process_USDA_match_to_flashtext,
    #     ),
    #     # To match the USDA like "V34626"
    #     OCRRegex(
    #         re.compile(r"[A-Z]\d{1,5}[A-Z]?"),
    #         field=OCRField.full_text_contiguous,
    #         lowercase=False,
    #         processing_func=process_USDA_match_to_flashtext,
    #     ),
    # ],
}


def find_packager_codes_regex(ocr_result: Union[OCRResult, str]) -> list[Prediction]:
    results: list[Prediction] = []

    for regex_code, regex_list in PACKAGER_CODE.items():
        for ocr_regex in regex_list:
            text = get_text(ocr_result, ocr_regex, ocr_regex.lowercase)

            if not text:
                continue

            for match in ocr_regex.regex.finditer(text):
                if ocr_regex.processing_func is None:
                    value = match.group(0)
                else:
                    value = ocr_regex.processing_func(match)

                if value is not None:
                    results.append(
                        Prediction(
                            value=value,
                            data={
                                "raw": match.group(0),
                                "type": regex_code,
                                "notify": ocr_regex.notify,
                            },
                            type=PredictionType.packager_code,
                            automatic_processing=True,
                        )
                    )

    return results


def generate_fishing_code_keyword_processor() -> KeywordProcessor:
    codes = text_file_iter(settings.OCR_FISHING_FLASHTEXT_DATA_PATH)
    return generate_keyword_processor(("{}||{}".format(c.upper(), c) for c in codes))


def extract_fishing_code(processor: KeywordProcessor, text: str) -> list[Prediction]:
    predictions = []

    for (key, _), span_start, span_end in processor.extract_keywords(
        text, span_info=True
    ):
        match_str = text[span_start:span_end]
        predictions.append(
            Prediction(
                type=PredictionType.packager_code,
                value=key,
                predictor="flashtext",
                data={"type": "fishing", "raw": match_str, "notify": False},
                automatic_processing=True,
            )
        )

    return predictions


FISHING_KEYWORD_PROCESSOR_STORE = CachedStore(
    fetch_func=generate_fishing_code_keyword_processor, expiration_interval=None
)


def find_packager_codes(ocr_result: Union[OCRResult, str]) -> list[Prediction]:
    predictions = find_packager_codes_regex(ocr_result)
    processor = FISHING_KEYWORD_PROCESSOR_STORE.get()
    text = get_text(ocr_result)
    predictions += extract_fishing_code(processor, text)
    return predictions



from base64 import b64decode
from urllib.parse import unquote


class BasicAuthDecodeError(Exception):
    pass


def basic_decode(encoded_str: str) -> tuple[str, str]:
    """Decode an encrypted HTTP basic authentication string. Returns a tuple of
    the form (username, password), and raises a BasicAuthDecodeError exception if
    nothing could be decoded.
    """
    split = encoded_str.strip().split(" ")

    # If split is only one element, try to decode the username and password
    # directly.
    if len(split) == 1:
        try:
            username, password = b64decode(split[0]).decode().split(":", 1)
        except Exception:
            raise BasicAuthDecodeError()

    # If there are only two elements, check the first and ensure it says
    # 'basic' so that we know we're about to decode the right thing. If not,
    # bail out.
    elif len(split) == 2:
        if split[0].strip().lower() == "basic":
            try:
                username, password = b64decode(split[1]).decode().split(":", 1)
            except Exception:
                raise BasicAuthDecodeError()
        else:
            raise BasicAuthDecodeError()

    # If there are more than 2 elements, something crazy must be happening.
    # Bail.
    else:
        raise BasicAuthDecodeError()

    return unquote(username), unquote(password)



import operator
import re
from dataclasses import InitVar, asdict, dataclass, field
from typing import Iterable, Optional

from robotoff.prediction.langid import DEFAULT_LANGUAGE_IDENTIFIER, LanguageIdentifier
from robotoff.spellcheck.utils import FR_KNOWN_TOKENS_CACHE
from robotoff.utils.text import get_blank_nlp

LANGUAGE_ALLOWED = "fr"
LANGUAGE_IDENTIFIER: LanguageIdentifier = DEFAULT_LANGUAGE_IDENTIFIER.get()

# Food additives (EXXX) may be mistaken from one another, because of their edit distance proximity
BLACKLIST_RE = re.compile(r"(?:\d+(?:[,.]\d+)?\s*%)|(?:[0-9])(?![\w-])")
PUNCTUATION_BLACKLIST_RE = re.compile(r"[_•:]")
E_BLACKLIST_RE = re.compile(r"(?<!\w)(?:E ?\d{3,5}[a-z]*)")
SPLITTER_CHAR = {"(", ")", ",", ";", "[", "]", "-", "{", "}"}


@dataclass
class Offset:
    start: int
    end: int

    def __lt__(self, other):
        return self.start < other.start


@dataclass
class AtomicCorrection:
    original: str
    correction: str
    offset: Offset
    force_valid: bool = False
    score: Optional[int] = None
    model: Optional[str] = None

    def is_valid(self, plural: bool = True, original_known: bool = True) -> bool:
        if self.force_valid:
            return True

        if plural and self._is_plural():
            return False

        if original_known and self._is_original_known():
            return False

        # Tokens with numbers are tricky to correct
        if any(x.isdigit() for x in self.correction):
            return False

        return True

    def has_difference(self) -> bool:
        return self.correction != self.original

    def _is_plural(self) -> bool:
        original_str = self.original.lower()
        correction_str = self.correction.lower()
        return (
            original_str.endswith("s")
            and correction_str == original_str[:-1]
            or correction_str.endswith("s")
            and original_str == correction_str[:-1]
        )

    def _is_original_known(self) -> bool:
        nlp = get_blank_nlp("fr")
        known_tokens = FR_KNOWN_TOKENS_CACHE.get()

        for token in nlp(self.original):
            if token.lower_ not in known_tokens:
                return False
        return True


@dataclass
class SpellcheckIteration:
    original: str
    model: Optional[str] = None
    correction: InitVar[Optional[str]] = None
    atomic_corrections: list[AtomicCorrection] = field(default_factory=list)

    def __post_init__(self, correction: str):
        if correction is not None:
            self.add_full_correction(correction)

    @property
    def corrected_text(self) -> str:
        valid_atomic_corrections = [
            correction
            for correction in self.atomic_corrections
            if correction.is_valid()
        ]

        if not valid_atomic_corrections:
            return self.original

        sorted_atomic_corrections = sorted(
            valid_atomic_corrections,
            key=operator.attrgetter("offset"),
        )

        last_correction = None
        corrected_fragments = []
        for atomic_correction in sorted_atomic_corrections:
            if last_correction is None:
                corrected_fragments.append(
                    self.original[: atomic_correction.offset.start]
                )
            else:
                corrected_fragments.append(
                    self.original[
                        last_correction.offset.end : atomic_correction.offset.start
                    ]
                )
            corrected_fragments.append(atomic_correction.correction)
            last_correction = atomic_correction

        if last_correction is not None:
            corrected_fragments.append(self.original[last_correction.offset.end :])

        return "".join(corrected_fragments)

    def add_atomic_correction(
        self, correction: str, offset: Offset, score: int
    ) -> None:
        self.atomic_corrections.append(
            AtomicCorrection(
                original=self.original[offset.start : offset.end],
                correction=correction,
                offset=offset,
                score=score,
                model=self.model,
            )
        )

    def add_full_correction(self, correction: str) -> None:
        assert len(self.atomic_corrections) == 0
        self.atomic_corrections = [
            AtomicCorrection(
                original=self.original,
                correction=correction,
                offset=Offset(0, len(self.original)),
                force_valid=True,
                model=self.model,
            )
        ]


class SpellcheckItem:
    def __init__(self, original: str):
        self.iterations: list[SpellcheckIteration] = []
        self.original: str = original
        self.is_lang_allowed: bool = self.__is_lang_allowed()

    @property
    def latest_correction(self) -> str:
        if len(self.iterations) > 0:
            return self.iterations[-1].corrected_text
        else:
            return self.original

    @property
    def corrections(self) -> list[dict]:
        return [
            dict(asdict(atomic_correction), is_valid=atomic_correction.is_valid())
            for atomic_correction in self.all_atomic_corrections
            if atomic_correction.has_difference()
        ]

    @property
    def all_atomic_corrections(self) -> list[AtomicCorrection]:
        return [
            atomic_correction
            for iteration in self.iterations
            for atomic_correction in iteration.atomic_corrections
        ]

    def __is_lang_allowed(self) -> bool:
        languages = LANGUAGE_IDENTIFIER.predict(self.original.lower(), threshold=0.5)
        if len(languages) == 0:
            return True
        return languages[0].lang == LANGUAGE_ALLOWED

    def update_correction(self, correction: str, model: str = "UNK"):
        self.iterations.append(
            SpellcheckIteration(
                original=self.latest_correction,
                correction=correction,
                model=model,
            )
        )


@dataclass
class Ingredients:
    text: str
    normalized_text: str
    offsets: list[Offset] = field(default_factory=list)

    def __iter__(self) -> Iterable[str]:
        yield from self.get_iter()

    def get_iter(self) -> Iterable[str]:
        for index, _ in enumerate(self.offsets):
            yield self.get_normalized_ingredient_text(index)

    def get_ingredient_text(self, index: int) -> str:
        offset = self.offsets[index]
        return self.text[offset.start : offset.end]

    def get_normalized_ingredient_text(self, index: int) -> str:
        offset = self.offsets[index]
        return self.normalized_text[offset.start : offset.end]

    def count(self) -> int:
        return len(self.offsets)

    @classmethod
    def from_text(cls, text: str, remove_blacklist: bool = True):
        if remove_blacklist:
            normalized_text = cls.process_remove_blacklist(text)
        else:
            normalized_text = text

        offsets = []
        chars = []
        start_idx = 0

        for idx, char in enumerate(normalized_text):
            if char in SPLITTER_CHAR:
                offsets.append(Offset(start_idx, idx))
                start_idx = idx + 1
                chars.append(" ")
            else:
                chars.append(char)

        if start_idx != len(normalized_text):
            offsets.append(Offset(start_idx, len(normalized_text)))

        normalized_text = "".join(chars)
        return cls(text, normalized_text, offsets)

    @classmethod
    def process_remove_blacklist(cls, text: str) -> str:
        text_without_blacklist = text

        for regex in (E_BLACKLIST_RE, BLACKLIST_RE, PUNCTUATION_BLACKLIST_RE):
            while True:
                try:
                    match = next(regex.finditer(text_without_blacklist))
                except StopIteration:
                    break
                if match:
                    start = match.start()
                    end = match.end()
                    text_without_blacklist = (
                        text_without_blacklist[:start]
                        + " " * (end - start)
                        + text_without_blacklist[end:]
                    )
                else:
                    break

        return text_without_blacklist



"""Interacting with OFF server to eg. update products or get infos
"""
import enum
import re
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse

from robotoff import settings
from robotoff.utils import get_logger, http_session

logger = get_logger(__name__)


class OFFAuthentication:
    def __init__(
        self,
        session_cookie: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ):
        if session_cookie is None and username is None and password is None:
            raise ValueError(
                "one of session_cookie or credentials (username and password) must be provided"
            )

        self.session_cookie = session_cookie
        self.username = username
        self.password = password

    def __eq__(self, other):
        """equality - we may use it in tests"""
        return (
            self.username == other.username
            and self.password == other.password
            and self.session_cookie == other.session_cookie
        )

    def get_username(self) -> Optional[str]:
        if self.username is not None:
            return self.username

        elif self.session_cookie is not None:
            splitted = self.session_cookie.split("&")

            if splitted:
                is_next = False
                for split in splitted:
                    if split == "user_id":
                        is_next = True
                        continue
                    elif is_next:
                        if split:
                            return split
                        else:
                            break

            logger.warning(
                "Unable to extract username from session cookie: %s",
                self.session_cookie,
            )

        return None


class ServerType(enum.Enum):
    off = 1
    obf = 2
    opff = 3
    opf = 4


API_URLS: dict[ServerType, str] = {
    ServerType.off: settings.BaseURLProvider().get(),
    ServerType.obf: "https://world.openbeautyfacts.org",
    ServerType.opf: "https://world.openproductfacts.org",
    ServerType.opff: "https://world.openpetfoodfacts.org",
}


BARCODE_PATH_REGEX = re.compile(r"^(...)(...)(...)(.*)$")


def get_source_from_url(ocr_url: str) -> str:
    url_path = urlparse(ocr_url).path

    if url_path.startswith("/images/products"):
        url_path = url_path[len("/images/products") :]

    if url_path.endswith(".json"):
        url_path = str(Path(url_path).with_suffix(".jpg"))

    return url_path


def get_barcode_from_url(url: str) -> Optional[str]:
    url_path = urlparse(url).path
    return get_barcode_from_path(url_path)


def get_barcode_from_path(path: str) -> Optional[str]:
    barcode = ""

    for parent in Path(path).parents:
        if parent.name.isdigit():
            barcode = parent.name + barcode
        else:
            break

    return barcode or None


def get_product_update_url(server: Union[ServerType, str]) -> str:
    return "{}/cgi/product_jqm2.pl".format(get_base_url(server))


def get_product_image_select_url(server: Union[ServerType, str]) -> str:
    return "{}/cgi/product_image_crop.pl".format(get_base_url(server))


def get_api_product_url(server: Union[ServerType, str]) -> str:
    return "{}/api/v0/product".format(get_base_url(server))


def get_base_url(server: Union[ServerType, str]) -> str:
    if isinstance(server, str):
        server = server.replace("api", "world")
        # get scheme, https on prod, but http in dev
        scheme = settings.BaseURLProvider().scheme
        return f"{scheme}://{server}"
    else:
        if server not in API_URLS:
            raise ValueError("unsupported server type: {}".format(server))

        return API_URLS[server]


def get_server_type(server_domain: str) -> ServerType:
    """Return the server type (off, obf, opff, opf) associated with the server
    domain, or None if the server_domain was not recognized."""
    server_split = server_domain.split(".")

    if len(server_split) == 3:
        subdomain, domain, tld = server_split

        if domain == "openfoodfacts":
            return ServerType.off
        elif domain == "openbeautyfacts":
            return ServerType.obf
        elif domain == "openpetfoodfacts":
            return ServerType.opff
        elif domain == "openproductsfacts":
            return ServerType.opf

    raise ValueError("unknown server domain: {}".format(server_domain))


def split_barcode(barcode: str) -> list[str]:
    if not barcode.isdigit():
        raise ValueError("unknown barcode format: {}".format(barcode))

    match = BARCODE_PATH_REGEX.fullmatch(barcode)

    if match:
        return [x for x in match.groups() if x]

    return [barcode]


def generate_image_path(barcode: str, image_id: str) -> str:
    splitted_barcode = split_barcode(barcode)
    return "/{}/{}.jpg".format("/".join(splitted_barcode), image_id)


def generate_json_path(barcode: str, image_id: str) -> str:
    splitted_barcode = split_barcode(barcode)
    return "/{}/{}.json".format("/".join(splitted_barcode), image_id)


def generate_json_ocr_url(barcode: str, image_id: str) -> str:
    return settings.OFF_IMAGE_BASE_URL + generate_json_path(barcode, image_id)


def generate_image_url(barcode: str, image_id: str) -> str:
    return settings.OFF_IMAGE_BASE_URL + generate_image_path(barcode, image_id)


def is_valid_image(barcode: str, image_id: str) -> bool:
    product = get_product(barcode, fields=["images"])

    if product is None:
        return False

    images = product.get("images", {})

    return image_id in images


def get_product(
    barcode: str,
    fields: Optional[list[str]] = None,
    server: Optional[Union[ServerType, str]] = None,
    timeout: Optional[int] = 10,
) -> Optional[dict]:
    fields = fields or []

    if server is None:
        server = ServerType.off

    url = get_api_product_url(server) + "/{}.json".format(barcode)

    if fields:
        # requests escape comma in URLs, as expected, but openfoodfacts server
        # does not recognize escaped commas.
        # See https://github.com/openfoodfacts/openfoodfacts-server/issues/1607
        url += "?fields={}".format(",".join(fields))

    r = http_session.get(url, timeout=timeout)

    if r.status_code != 200:
        return None

    data = r.json()

    if data["status_verbose"] != "product found":
        return None

    return data["product"]


def add_category(
    barcode: str, category: str, insight_id: Optional[str] = None, **kwargs
):
    comment = "[robotoff] Adding category '{}'".format(category)

    if insight_id:
        comment += ", ID: {}".format(insight_id)

    params = {
        "code": barcode,
        "add_categories": category,
        "comment": comment,
    }
    update_product(params, **kwargs)


def update_quantity(
    barcode: str, quantity: str, insight_id: Optional[str] = None, **kwargs
):
    comment = "[robotoff] Updating quantity to '{}'".format(quantity)

    if insight_id:
        comment += ", ID: {}".format(insight_id)

    params = {
        "code": barcode,
        "quantity": quantity,
        "comment": comment,
    }
    update_product(params, **kwargs)


def update_emb_codes(
    barcode: str, emb_codes: list[str], insight_id: Optional[str] = None, **kwargs
):
    emb_codes_str = ",".join(emb_codes)

    comment = "[robotoff] Adding packager codes '{}'".format(emb_codes_str)

    if insight_id:
        comment += ", ID: {}".format(insight_id)

    params = {
        "code": barcode,
        "emb_codes": emb_codes_str,
        "comment": comment,
    }
    update_product(params, **kwargs)


def update_expiration_date(
    barcode: str, expiration_date: str, insight_id: Optional[str] = None, **kwargs
):
    comment = "[robotoff] Adding expiration date '{}'".format(expiration_date)

    if insight_id:
        comment += ", ID: {}".format(insight_id)

    params = {
        "code": barcode,
        "expiration_date": expiration_date,
        "comment": comment,
    }
    update_product(params, **kwargs)


def add_label_tag(
    barcode: str, label_tag: str, insight_id: Optional[str] = None, **kwargs
):
    comment = "[robotoff] Adding label tag '{}'".format(label_tag)

    if insight_id:
        comment += ", ID: {}".format(insight_id)

    params = {
        "code": barcode,
        "add_labels": label_tag,
        "comment": comment,
    }
    update_product(params, **kwargs)


def add_brand(barcode: str, brand: str, insight_id: Optional[str] = None, **kwargs):
    comment = "[robotoff] Adding brand '{}'".format(brand)

    if insight_id:
        comment += ", ID: {}".format(insight_id)

    params = {
        "code": barcode,
        "add_brands": brand,
        "comment": comment,
    }
    update_product(params, **kwargs)


def add_store(barcode: str, store: str, insight_id: Optional[str] = None, **kwargs):
    comment = "[robotoff] Adding store '{}'".format(store)

    if insight_id:
        comment += ", ID: {}".format(insight_id)

    params = {
        "code": barcode,
        "add_stores": store,
        "comment": comment,
    }
    update_product(params, **kwargs)


def add_packaging(
    barcode: str, packaging: str, insight_id: Optional[str] = None, **kwargs
):
    comment = "[robotoff] Adding packaging '{}'".format(packaging)

    if insight_id:
        comment += ", ID: {}".format(insight_id)

    params = {
        "code": barcode,
        "add_packaging": packaging,
        "comment": comment,
    }
    update_product(params, **kwargs)


def save_ingredients(
    barcode: str,
    ingredient_text: str,
    insight_id: Optional[str] = None,
    lang: Optional[str] = None,
    comment: Optional[str] = None,
    **kwargs,
):
    ingredient_key = "ingredients_text" if lang is None else f"ingredients_text_{lang}"

    if comment:
        comment = "[robotoff] Ingredient spellcheck correction ({})".format(comment)
    else:
        comment = "[robotoff] Ingredient spellcheck correction"

    if insight_id:
        comment += ", ID: {}".format(insight_id)

    params = {
        "code": barcode,
        "comment": comment,
        ingredient_key: ingredient_text,
    }
    update_product(params, **kwargs)


def update_product(
    params: dict,
    server_domain: Optional[str] = None,
    auth: Optional[OFFAuthentication] = None,
    timeout: Optional[int] = 15,
):
    if server_domain is None:
        server_domain = settings.OFF_SERVER_DOMAIN

    url = get_product_update_url(server_domain)

    comment = params.get("comment")
    cookies = None

    if auth is not None:
        if auth.session_cookie:
            cookies = {
                "session": auth.session_cookie,
            }
        elif auth.username:
            params["user_id"] = auth.username
            params["password"] = auth.password
    else:
        params.update(settings.off_credentials())

        if comment:
            params["comment"] = comment + " (automated edit)"

    if cookies is None and not params.get("password"):
        raise ValueError(
            "a password or a session cookie is required to update a product"
        )
    r = http_session.get(
        url,
        params=params,
        auth=settings._off_request_auth,
        cookies=cookies,
        timeout=timeout,
    )

    r.raise_for_status()
    json = r.json()

    status = json.get("status_verbose")

    if status != "fields saved":
        logger.warning("Unexpected status during product update: %s", status)


def move_to(barcode: str, to: ServerType, timeout: Optional[int] = 10) -> bool:
    if get_product(barcode, server=to) is not None:
        return False

    url = "{}/cgi/product_jqm.pl".format(settings.BaseURLProvider().get())
    params = {
        "type": "edit",
        "code": barcode,
        "new_code": str(to),
        **settings.off_credentials(),
    }
    r = http_session.get(url, params=params, timeout=timeout)
    data = r.json()
    return data["status"] == 1


def select_rotate_image(
    barcode: str,
    image_id: str,
    image_key: Optional[str] = None,
    rotate: Optional[int] = None,
    server_domain: Optional[str] = None,
    auth: Optional[OFFAuthentication] = None,
    timeout: Optional[int] = 15,
):
    if server_domain is None:
        server_domain = settings.OFF_SERVER_DOMAIN

    url = get_product_image_select_url(server_domain)
    cookies = None
    params = {
        "code": barcode,
        "imgid": image_id,
    }

    if rotate is not None:
        if rotate not in (90, 180, 270):
            raise ValueError("invalid value for rotation angle: {}".format(rotate))

        params["angle"] = str(rotate)

    if image_key is not None:
        params["id"] = image_key

    if auth is not None:
        if auth.session_cookie:
            cookies = {
                "session": auth.session_cookie,
            }
        elif auth.username and auth.password:
            params["user_id"] = auth.username
            params["password"] = auth.password
    else:
        params.update(settings.off_credentials())

    if cookies is None and not params.get("password"):
        raise ValueError(
            "a password or a session cookie is required to select an image"
        )

    r = http_session.post(
        url,
        data=params,
        auth=settings._off_request_auth,
        cookies=cookies,
        timeout=timeout,
    )

    r.raise_for_status()
    return r


def normalize_tag(value, lowercase=True):
    """given a value normalize it to a tag (as in taxonomies)

    This means removing accents, lowercasing, replacing spaces with dashes, etc..
    """
    # removing accents
    value = re.sub(r"[¢£¤¥§©ª®°²³µ¶¹º¼½¾×‰€™]", "-", value)
    value = re.sub(r"[éè]", "e", value)
    value = re.sub(r"[à]", "a", value)
    value = re.sub(r"[ù]", "u", value)
    # changing unwanted character to "-"
    value = re.sub(r"&\w+;", "-", value)
    value = re.sub(
        r"[\s!\"#\$%&'()*+,\/:;<=>?@\[\\\]^_`{\|}~¡¢£¤¥¦§¨©ª«¬®¯°±²³´µ¶·¸¹º»¼½¾¿×ˆ˜–—‘’‚“”„†‡•…‰‹›€™\t]",  # noqa: E501
        "-",
        value,
    )
    # lowering the value if wanted
    if lowercase:
        value = value.lower()
    # removing excess "-"
    value = re.sub(r"-+", "-", value)
    value = value.strip("-")
    return value



import copy
import datetime
import logging
import os
from pathlib import Path
from typing import Optional

import sentry_sdk
from sentry_sdk.integrations import Integration
from sentry_sdk.integrations.logging import LoggingIntegration

# Robotoff instance gives the environment, either `prod` or `dev`
# (`dev` by default).
# If `prod` is used, openfoodfacts.org domain will be used by default,
# and openfoodfacts.net if `dev` is used.
# Messages to Slack are only enabled if `ROBOTOFF_INSTANCE=prod`.
ROBOTOFF_INSTANCE = os.environ.get("ROBOTOFF_INSTANCE", "dev")


# Returns the top-level-domain (TLD) for the Robotoff instance.
def _instance_tld() -> str:
    if ROBOTOFF_INSTANCE == "prod":
        return "org"
    elif ROBOTOFF_INSTANCE == "dev":
        return "net"
    else:
        return ROBOTOFF_INSTANCE


_default_robotoff_domain = f"openfoodfacts.{_instance_tld()}"

# `ROBOTOFF_DOMAIN` can be used to overwrite the Product Opener domain used.
# If empty, the domain will be inferred from `ROBOTOFF_INSTANCE`
_robotoff_domain = os.environ.get("ROBOTOFF_DOMAIN", _default_robotoff_domain)


class BaseURLProvider(object):
    """BaseURLProvider allows to fetch a base URL for Product Opener/Robotoff.

    Example usage: BaseURLProvider().robotoff().get() returns the Robotoff URL.
    """

    def __init__(self):
        self.domain = os.environ.get(
            "ROBOTOFF_DOMAIN", "openfoodfacts.%s" % _instance_tld()
        )
        self.url = "%(scheme)s://%(prefix)s.%(domain)s"
        self.prefix = "world"
        self.scheme = os.environ.get("ROBOTOFF_SCHEME", "https")

    def clone(self):
        return copy.deepcopy(self)

    def robotoff(self):
        result = self.clone()
        result.prefix = "robotoff"
        return result

    def static(self):
        result = self.clone()
        result.prefix = "static"
        # locally we may want to change it, give environment a chance
        static_domain = os.environ.get("STATIC_OFF_DOMAIN", "")
        if static_domain:
            if "://" in static_domain:
                result.scheme, static_domain = static_domain.split("://", 1)
            result.domain = static_domain
        return result

    def country(self, country_code: str):
        result = self.clone()
        result.prefix = country_code
        return result

    def get(self):
        return self.url % {
            "scheme": self.scheme,
            "prefix": self.prefix,
            "domain": self.domain,
        }


PROJECT_DIR = Path(__file__).parent.parent
DATA_DIR = PROJECT_DIR / "data"
DATASET_DIR = PROJECT_DIR / "datasets"
DATASET_DIR.mkdir(exist_ok=True)
I18N_DIR = PROJECT_DIR / "i18n"
LABEL_LOGOS_PATH = DATA_DIR / "label_logos.json"
JSONL_DATASET_PATH = DATASET_DIR / "products.jsonl.gz"
JSONL_DATASET_ETAG_PATH = DATASET_DIR / "products-etag.txt"
JSONL_MIN_DATASET_PATH = DATASET_DIR / "products-min.jsonl.gz"
DATASET_CHECK_MIN_PRODUCT_COUNT = 1000000

# Products JSONL

JSONL_DATASET_URL = (
    BaseURLProvider().static().get() + "/data/openfoodfacts-products.jsonl.gz"
)

TAXONOMY_URLS = {
    "category": BaseURLProvider().static().get()
    + "/data/taxonomies/categories.full.json",
    "ingredient": BaseURLProvider().static().get()
    + "/data/taxonomies/ingredients.full.json",
    "label": BaseURLProvider().static().get() + "/data/taxonomies/labels.full.json",
    "brand": BaseURLProvider().static().get() + "/data/taxonomies/brands.full.json",
    "packaging_shape": BaseURLProvider().static().get()
    + "/data/taxonomies/packaging_shapes.full.json",
    "packaging_material": BaseURLProvider().static().get()
    + "/data/taxonomies/packaging_materials.full.json",
    "packaging_recycling": BaseURLProvider().static().get()
    + "/data/taxonomies/packaging_recycling.full.json",
}

OFF_IMAGE_BASE_URL = BaseURLProvider().static().get() + "/images/products"

_off_password = os.environ.get("OFF_PASSWORD", "")
_off_user = os.environ.get("OFF_USER", "")
_off_request_auth = ("off", "off") if _instance_tld() == "net" else None


def off_credentials() -> dict[str, str]:
    return {"user_id": _off_user, "password": _off_password}


OFF_SERVER_DOMAIN = "api." + BaseURLProvider().domain
EVENTS_API_URL = os.environ.get(
    "EVENTS_API_URL", "https://events." + BaseURLProvider().domain
)

CATEGORY_MATCHER_DIR = DATA_DIR / "category_matcher"
CATEGORY_MATCHER_MATCH_MAPS = {
    "category": CATEGORY_MATCHER_DIR / "category_match_maps.json.gz",
    "ingredient": CATEGORY_MATCHER_DIR / "ingredient_match_maps.json.gz",
}
CATEGORY_MATCHER_INTERSECT = (
    CATEGORY_MATCHER_DIR / "category_ingredient_intersect.json.gz"
)

# Taxonomies are huge JSON files that describe many concepts in OFF, in many languages, with synonyms. Those are the full version of taxos.

TAXONOMY_DIR = DATA_DIR / "taxonomies"
TAXONOMY_PATHS = {
    "category": TAXONOMY_DIR / "categories.full.json.gz",
    "ingredient": TAXONOMY_DIR / "ingredients.full.json.gz",
    "label": TAXONOMY_DIR / "labels.full.json.gz",
    "brand": TAXONOMY_DIR / "brands.full.json.gz",
}
INGREDIENTS_FR_PATH = TAXONOMY_DIR / "ingredients_fr.txt"
INGREDIENT_TOKENS_PATH = TAXONOMY_DIR / "ingredients_tokens.txt"
FR_TOKENS_PATH = TAXONOMY_DIR / "fr_tokens_lower.gz"

# Spellchecking parameters. Wauplin and Raphael are the experts.

SPELLCHECK_DIR = DATA_DIR / "spellcheck"
SPELLCHECK_PATTERNS_PATHS = {
    "fr": SPELLCHECK_DIR / "patterns_fr.txt",
}

# Credentials for the Robotoff insights database

POSTGRES_HOST = os.environ.get("POSTGRES_HOST", "localhost")
POSTGRES_DB = os.environ.get("POSTGRES_DB", "postgres")
POSTGRES_USER = os.environ.get("POSTGRES_USER", "postgres")
POSTGRES_PASSWORD = os.environ.get("POSTGRES_PASSWORD", "postgres")

# Mongo used to be on the same server as Robotoff

MONGO_URI = os.environ.get("MONGO_URI", "mongodb://mongodb:27017")

# Redis
REDIS_HOST = os.environ.get("REDIS_HOST", "localhost")

# how many seconds should we wait to compute insight on product updated
UPDATED_PRODUCT_WAIT = float(os.environ.get("ROBOTOFF_UPDATED_PRODUCT_WAIT", 10))

# Elastic Search is used for simple category prediction and spellchecking.

ELASTICSEARCH_HOSTS = os.environ.get("ELASTICSEARCH_HOSTS", "localhost:9200").split(",")
ELASTICSEARCH_TYPE = "document"


class ElasticsearchIndex:
    PRODUCT = "product"

    SUPPORTED_INDICES = {
        PRODUCT: (PROJECT_DIR / "robotoff/elasticsearch/index/product_index.json"),
    }


# image moderation service
IMAGE_MODERATION_SERVICE_URL: Optional[str] = os.environ.get(
    "IMAGE_MODERATION_SERVICE_URL", None
)

# Slack paramaters for notifications about detection
_slack_token = os.environ.get("SLACK_TOKEN", "")


# Returns the slack token to use for posting alerts if the current instance is the 'prod' instance.
# For all other instances, the empty string is returned.
def slack_token() -> str:
    if ROBOTOFF_INSTANCE == "prod":
        if _slack_token != "":
            return _slack_token
        else:
            raise ValueError("No SLACK_TOKEN specified for prod Robotoff")

    if _slack_token != "":
        raise ValueError("SLACK_TOKEN specified for non-prod Robotoff")
    return ""


# Sentry for error reporting
_sentry_dsn = os.environ.get("SENTRY_DSN")


def init_sentry(integrations: Optional[list[Integration]] = None):
    if _sentry_dsn:
        integrations = integrations or []
        integrations.append(
            LoggingIntegration(
                level=logging.INFO,  # Capture info and above as breadcrumbs
                event_level=logging.WARNING,  # Send warning and errors as events
            )
        )
        sentry_sdk.init(  # type:ignore # mypy say it's abstract
            _sentry_dsn,
            environment=ROBOTOFF_INSTANCE,
            integrations=integrations,
        )
    elif ROBOTOFF_INSTANCE == "prod":
        raise ValueError("No SENTRY_DSN specified for prod Robotoff")


OCR_DATA_DIR = DATA_DIR / "ocr"
OCR_BRANDS_PATH = OCR_DATA_DIR / "brand.txt"
OCR_TAXONOMY_BRANDS_PATH = OCR_DATA_DIR / "brand_from_taxonomy.gz"
OCR_LOGO_ANNOTATION_BRANDS_DATA_PATH = OCR_DATA_DIR / "brand_logo_annotation.txt"
OCR_STORES_DATA_PATH = OCR_DATA_DIR / "store_regex.txt"
OCR_STORES_NOTIFY_DATA_PATH = OCR_DATA_DIR / "store_notify.txt"
OCR_LOGO_ANNOTATION_LABELS_DATA_PATH = OCR_DATA_DIR / "label_logo_annotation.txt"
OCR_LABEL_FLASHTEXT_DATA_PATH = OCR_DATA_DIR / "label_flashtext.txt"
OCR_USDA_CODE_FLASHTEXT_DATA_PATH = OCR_DATA_DIR / "USDA_code_flashtext.txt"
OCR_LABEL_WHITELIST_DATA_PATH = OCR_DATA_DIR / "label_whitelist.txt"
# Try to detect MSC codes
OCR_FISHING_FLASHTEXT_DATA_PATH = OCR_DATA_DIR / "fishing_flashtext.txt"
OCR_TAXONOMY_BRANDS_BLACKLIST_PATH = OCR_DATA_DIR / "brand_taxonomy_blacklist.txt"
# Try to detect cosmetics in OFF
OCR_IMAGE_FLAG_BEAUTY_PATH = OCR_DATA_DIR / "image_flag_beauty.txt"
OCR_IMAGE_FLAG_MISCELLANEOUS_PATH = OCR_DATA_DIR / "image_flag_miscellaneous.txt"
OCR_PACKAGING_DATA_PATH = OCR_DATA_DIR / "packaging.txt"
OCR_TRACE_ALLERGEN_DATA_PATH = OCR_DATA_DIR / "trace_allergen.txt"
# Try to detect postal codes in France
OCR_CITIES_FR_PATH = OCR_DATA_DIR / "cities_laposte_hexasmal.json.gz"

BRAND_PREFIX_PATH = DATA_DIR / "brand_prefix.json.gz"

# When we're making queries to the API, so that we're not blocked by error
ROBOTOFF_USER_AGENT = "Robotoff Live Analysis"
# Models and ML

MODELS_DIR = PROJECT_DIR / "models"

# Tensorflow Serving host parameters

_tf_serving_host = os.environ.get("TF_SERVING_HOST", "localhost")
_tf_serving_http_port = os.environ.get("TF_SERVING_PORT", "8501")
TF_SERVING_BASE_URL = f"http://{_tf_serving_host}:{_tf_serving_http_port}/v1/models"


_triton_host = os.environ.get("TRITON_HOST", "localhost")
_triton_grpc_port = os.environ.get("TRITON_PORT", "8001")
TRITON_URI = f"{_triton_host}:{_triton_grpc_port}"

MODELS_DIR = PROJECT_DIR / "models"
OBJECT_DETECTION_IMAGE_MAX_SIZE = (1024, 1024)


# We require a minimum of 15 occurences of the brands already on OFF to perform the extraction. This reduces false positive.
# We require a minimum of 4 characters for the brand

BRAND_MATCHING_MIN_LENGTH = 4
BRAND_MATCHING_MIN_COUNT = 15

INFLUXDB_HOST = os.environ.get("INFLUXDB_HOST", "localhost")
INFLUXDB_PORT = int(os.environ.get("INFLUXDB_PORT", "8086"))
INFLUXDB_BUCKET = os.environ.get("INFLUXDB_BUCKET", "off_metrics")
INFLUXDB_AUTH_TOKEN = os.environ.get("INFLUXDB_AUTH_TOKEN")
INFLUXDB_ORG = os.environ.get("INFLUXDB_ORG", "off")

TEST_DIR = PROJECT_DIR / "tests"
TEST_DATA_DIR = TEST_DIR / "unit/data"

# Maximum interval between the upload of the insight image and the upload of
# the most recent image of the product to consider the insight image as
# trustworthy. Insights with untrustworthy images cannot be applied
# automatically.
# For example if days=120, it means that an insight based on an image that is
# more than 120 days older than the most recent product image cannot be applied
# automatically.
IMAGE_MAX_TIMEDELTA = datetime.timedelta(days=120)

# Number of minutes to wait before processing an insight automatically
INSIGHT_AUTOMATIC_PROCESSING_WAIT = int(
    os.environ.get("INSIGHT_AUTOMATIC_PROCESSING_WAIT", 10)
)



import functools
import re
from typing import Optional, Union

import pint

from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType
from robotoff.utils import get_logger

from .dataclass import OCRField, OCRRegex, OCRResult, get_text

logger = get_logger(__name__)


ureg = pint.UnitRegistry()


def normalize_weight(value: str, unit: str) -> tuple[float, str]:
    """Normalize the product weight unit to g for mass and mL for volumes.
    Return a (value, unit) tuple, where value is the normalized value as a
    float and unit either 'g' or 'ml'."""
    if "," in value:
        # pint does not recognize ',' separator
        value = value.replace(",", ".")

    if unit == "fl oz":
        # For nutrition labeling, a fluid ounce is equal to 30 ml
        value = str(float(value) * 30)
        unit = "ml"

    quantity = ureg.parse_expression("{} {}".format(value, unit))

    if ureg.gram in quantity.compatible_units():
        normalized_quantity = quantity.to(ureg.gram)
        normalized_unit = "g"
    elif ureg.liter in quantity.compatible_units():
        normalized_quantity = quantity.to(ureg.milliliter)
        normalized_unit = "ml"
    else:
        raise ValueError("unknown unit: {}".format(quantity.u))

    return normalized_quantity.magnitude, normalized_unit


def is_valid_weight(weight_value: str) -> bool:
    """Weight values are considered invalid if one of the following rules
    is met:
    - value is not convertible to a float
    - value is negative
    - value is not an integer
    - value starts with a 0 and does not have a '.' or ',' in it
    """
    weight_value = weight_value.replace(",", ".")

    if weight_value.startswith("0") and "." not in weight_value:
        return False

    try:
        weight_value_float = float(weight_value)
    except ValueError:
        logger.warning("Weight value is not a float: {}" "".format(weight_value))
        return False

    if weight_value_float <= 0:
        logger.debug("Weight value is <= 0")
        return False

    if float(int(weight_value_float)) != weight_value_float:
        logger.info(
            "Weight value is not an integer ({}), "
            "returning non valid".format(weight_value)
        )
        return False

    return True


def is_extreme_weight(normalized_value: float, unit: str) -> bool:
    if unit == "g":
        # weights above 10 kg
        return normalized_value >= 10000 or normalized_value <= 10
    elif unit == "ml":
        # volumes above 10 l
        return normalized_value >= 10000 or normalized_value <= 10

    raise ValueError("invalid unit: {}, 'g', or 'ml' " "expected".format(unit))


def is_suspicious_weight(normalized_value: float, unit: str) -> bool:
    """Return True is the weight is suspicious, i.e is likely wrongly
    detected."""
    if is_extreme_weight(normalized_value, unit):
        return True

    if normalized_value > 1000:
        # weight value is above 1000 and
        # last digit is not 0
        # See https://github.com/openfoodfacts/robotoff/issues/43
        last_digit = str(int(normalized_value))[-1]

        if last_digit != "0":
            return True

    return False


def process_product_weight(
    match: re.Match,
    prompt: bool,
    automatic_processing: bool,
    ending_prompt: bool = False,
) -> Optional[dict]:
    raw = match.group()

    if prompt:
        if ending_prompt:
            value = match.group(1)
            unit = match.group(2)
            prompt_str = match.group(3)

        else:
            prompt_str = match.group(1)
            value = match.group(2)
            unit = match.group(3)
    else:
        prompt_str = None
        value = match.group(1)
        unit = match.group(2)

    if unit in ("dle", "cle", "mge", "mle", "ge", "kge", "le"):
        # When the e letter often comes after the weight unit, the
        # space is often not detected
        unit = unit[:-1]

    if not is_valid_weight(value):
        return None

    # Strip value from endpoint point: '525. g' -> '525 g'
    value = value.strip(".")

    text = "{} {}".format(value, unit)
    normalized_value, normalized_unit = normalize_weight(value, unit)

    if is_suspicious_weight(normalized_value, normalized_unit):
        # Don't process the prediction automatically if the value
        # is suspicious (very high, low,...)
        automatic_processing = False

    result = {
        "text": text,
        "raw": raw,
        "value": value,
        "unit": unit,
        "normalized_value": normalized_value,
        "normalized_unit": normalized_unit,
        "automatic_processing": automatic_processing,
    }

    if prompt_str is not None:
        result["prompt"] = prompt_str

    return result


def process_multi_packaging(match) -> Optional[dict]:
    raw = match.group()

    count = match.group(1)
    value = match.group(2)
    unit = match.group(3)

    if unit in ("dle", "cle", "mge", "mle", "ge", "kge", "le"):
        # When the e letter often comes after the weight unit, the
        # space is often not detected
        unit = unit[:-1]

    if not is_valid_weight(value):
        return None

    normalized_value, normalized_unit = normalize_weight(value, unit)
    text = "{} x {} {}".format(count, value, unit)
    result = {
        "text": text,
        "raw": raw,
        "value": value,
        "unit": unit,
        "count": count,
        "normalized_value": normalized_value,
        "normalized_unit": normalized_unit,
        # Don't process the prediction automatically if the value
        # is suspiciously high
        "automatic_processing": not is_suspicious_weight(
            normalized_value, normalized_unit
        ),
    }

    return result


PRODUCT_WEIGHT_REGEX: dict[str, OCRRegex] = {
    "with_mention": OCRRegex(
        re.compile(
            r"(?<![a-z])(poids|poids net [aà] l'emballage|poids net|poids net égoutté|masse nette|volume net total|net weight|net wt\.?|peso neto|peso liquido|netto[ -]?gewicht)\s?:?\s?([0-9]+[,.]?[0-9]*)\s?(fl oz|dle?|cle?|mge?|mle?|lbs|oz|ge?|kge?|le?)(?![a-z])"
        ),
        field=OCRField.full_text_contiguous,
        lowercase=True,
        processing_func=functools.partial(
            process_product_weight, prompt=True, automatic_processing=True
        ),
        priority=1,
    ),
    "with_ending_mention": OCRRegex(
        re.compile(
            r"(?<![a-z])([0-9]+[,.]?[0-9]*)\s?(fl oz|dle?|cle?|mge?|mle?|lbs|oz|ge?|kge?|le?)\s(net)(?![a-z])"
        ),
        field=OCRField.full_text_contiguous,
        lowercase=True,
        processing_func=functools.partial(
            process_product_weight,
            prompt=True,
            ending_prompt=True,
            automatic_processing=True,
        ),
        priority=1,
    ),
    "multi_packaging": OCRRegex(
        re.compile(
            r"(?<![a-z])(\d+)\s?x\s?([0-9]+[,.]?[0-9]*)\s?(fl oz|dle?|cle?|mge?|mle?|lbs|oz|ge?|kge?|le?)(?![a-z])"
        ),
        field=OCRField.full_text_contiguous,
        lowercase=True,
        processing_func=process_multi_packaging,
        priority=2,
    ),
    "no_mention": OCRRegex(
        re.compile(
            r"(?<![a-z])([0-9]+[,.]?[0-9]*)\s?(dle|cle|mge|mle|ge|kge)(?![a-z])"
        ),
        field=OCRField.full_text_contiguous,
        lowercase=True,
        processing_func=functools.partial(
            process_product_weight, prompt=False, automatic_processing=False
        ),
        priority=3,
    ),
}


def find_product_weight(content: Union[OCRResult, str]) -> list[Prediction]:
    results = []

    for type_, ocr_regex in PRODUCT_WEIGHT_REGEX.items():
        text = get_text(content, ocr_regex)

        if not text:
            continue

        for match in ocr_regex.regex.finditer(text):
            if ocr_regex.processing_func is None:
                continue

            result = ocr_regex.processing_func(match)

            if result is None:
                continue

            result["matcher_type"] = type_
            result["priority"] = ocr_regex.priority
            result["notify"] = ocr_regex.notify
            value = result.pop("text")
            results.append(
                Prediction(
                    value=value,
                    type=PredictionType.product_weight,
                    automatic_processing=result["automatic_processing"],
                    data=result,
                )
            )

    return results



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['robotoff',
 'robotoff.app',
 'robotoff.cli',
 'robotoff.elasticsearch',
 'robotoff.elasticsearch.product',
 'robotoff.insights',
 'robotoff.prediction',
 'robotoff.prediction.category',
 'robotoff.prediction.category.neural',
 'robotoff.prediction.object_detection',
 'robotoff.prediction.object_detection.utils',
 'robotoff.prediction.ocr',
 'robotoff.scheduler',
 'robotoff.spellcheck',
 'robotoff.spellcheck.elasticsearch',
 'robotoff.spellcheck.patterns',
 'robotoff.spellcheck.percentages',
 'robotoff.spellcheck.vocabulary',
 'robotoff.utils',
 'robotoff.workers',
 'robotoff.workers.tasks']

package_data = \
{'': ['*'], 'robotoff.elasticsearch': ['index/*']}

install_requires = \
['APScheduler>=3.8.1,<3.9.0',
 'Pillow>=9.3.0,<9.4.0',
 'Pint==0.9',
 'cachetools>=5.2.0,<6.0.0',
 'dacite>=1.6.0,<1.7.0',
 'elasticsearch>=7.16.3,<7.17.0',
 'falcon-cors>=1.1.7,<1.2.0',
 'falcon-multipart>=0.2.0,<0.3.0',
 'falcon>=2.0.0,<2.1.0',
 'flashtext>=2.7,<2.8',
 'gunicorn>=20.1.0,<20.2.0',
 'influxdb-client>=1.34.0,<1.35.0',
 'jsonschema>=4.4.0,<4.5.0',
 'langid>=1.1.6,<1.2.0',
 'matplotlib>=3.5.0,<3.6.0',
 'more-itertools>=8.9.0,<8.10.0',
 'numpy>=1.23.5,<1.24.0',
 'orjson>=3.8.2,<3.9.0',
 'peewee>=3.14.4,<3.15.0',
 'protobuf>=3.17.3,<3.18.0',
 'psycopg2-binary>=2.9.1,<2.10.0',
 'py-healthcheck>=1.10.1,<2.0.0',
 'pymongo>=3.12.0,<3.13.0',
 'python-redis-lock>=4.0.0,<4.1.0',
 'requests>=2.28.1,<2.29.0',
 'rq>=1.11.1,<1.12.0',
 'sentry-sdk[falcon]>=1.11.0,<1.12.0',
 'spacy-lookups-data>=1.0.3,<2.0.0',
 'spacy>=3.4.1,<3.5.0',
 'tritonclient[grpc]>=2.26.0,<3.0.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['robotoff-cli = robotoff.cli.main:main']}

setup_kwargs = {
    'name': 'robotoff',
    'version': '1.15.1',
    'description': 'Real-time and batch prediction service for Openfoodfacts.',
    'long_description': None,
    'author': 'Openfoodfacts Team',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import requests

from robotoff.insights.extraction import get_predictions_from_product_name
from robotoff.insights.importer import import_insights, refresh_insights
from robotoff.models import with_db
from robotoff.off import ServerType, get_server_type
from robotoff.prediction.category.matcher import predict as predict_category_matcher
from robotoff.prediction.category.neural.category_classifier import CategoryClassifier
from robotoff.products import get_product
from robotoff.redis import Lock, LockedResourceException
from robotoff.taxonomy import TaxonomyType, get_taxonomy
from robotoff.utils import get_logger
from robotoff.utils.types import JSONType

logger = get_logger(__name__)


@with_db
def update_insights_job(barcode: str, server_domain: str):
    """This job is triggered by the webhook API, when product information has
    been updated.

    When a product is updated, Robotoff will:

    1. Generate new predictions related to the product's category and name.
    2. Regenerate all insights from the product associated predictions.
    """
    logger.info("Running `update_insights` for product %s (%s)", barcode, server_domain)

    try:
        with Lock(
            name=f"robotoff:product_update_job:{barcode}", expire=300, timeout=10
        ):
            # We handle concurrency thanks to the lock as the task will fetch
            # product from MongoDB at the time it runs, it's not worth
            # reprocessing with another task arriving concurrently.
            # The expire is there only in case the lock is not released
            # (process killed)
            product_dict = get_product(barcode)

            if product_dict is None:
                logger.warning("Updated product does not exist: %s", barcode)
                return

            updated_product_predict_insights(barcode, product_dict, server_domain)
            logger.info("Refreshing insights...")
            imported = refresh_insights(barcode, server_domain)
            logger.info("%s insights created after refresh", imported)
    except LockedResourceException:
        logger.info(
            f"Couldn't acquire product_update lock, skipping product_update for product {barcode}"
        )


def add_category_insight(barcode: str, product: JSONType, server_domain: str) -> bool:
    """Predict categories for product and import predicted category insight.

    :param barcode: product barcode
    :param product: product as retrieved from application
    :param server_domain: the server the product belongs to
    :return: True if at least one category insight was imported
    """
    if get_server_type(server_domain) != ServerType.off:
        return False

    logger.info("Predicting product categories...")
    # predict category using matching algorithm on product name
    product_predictions = predict_category_matcher(product)

    # predict category using neural model
    try:
        product_predictions += CategoryClassifier(
            get_taxonomy(TaxonomyType.category.name)
        ).predict(product)
    except requests.exceptions.HTTPError as e:
        resp = e.response
        logger.error(
            f"Category classifier returned an error: {resp.status_code}: %s", resp.text
        )

    if len(product_predictions) < 1:
        return False

    for prediction in product_predictions:
        prediction.barcode = barcode

    imported = import_insights(product_predictions, server_domain)
    logger.info("%s category insight imported for product %s", imported, barcode)

    return bool(imported)


def updated_product_predict_insights(
    barcode: str, product: JSONType, server_domain: str
) -> bool:
    updated = add_category_insight(barcode, product, server_domain)
    product_name = product.get("product_name")

    if not product_name:
        return updated

    logger.info("Generating predictions from product name...")
    predictions_all = get_predictions_from_product_name(barcode, product_name)
    imported = import_insights(predictions_all, server_domain)
    logger.info("%s insights imported for product %s", imported, barcode)

    if imported:
        updated = True

    return updated



import functools
import re
import unicodedata

import spacy

from .fold_to_ascii import fold

CONSECUTIVE_SPACES_REGEX = re.compile(r" {2,}")


def strip_accents_ascii(s: str) -> str:
    """Transform accentuated unicode symbols into ascii or nothing

    Warning: this solution is only suited for languages that have a direct
    transliteration to ASCII symbols.

    Parameters
    ----------
    s : string
        The string to strip

    See also
    --------
    strip_accents_unicode
        Remove accentuated char for any unicode symbol.
    """
    nkfd_form = unicodedata.normalize("NFKD", s)
    return nkfd_form.encode("ASCII", "ignore").decode("ASCII")


def strip_accents_ascii_v2(s):
    return fold(s)


def strip_consecutive_spaces(text: str) -> str:
    """Convert a sequence of 2+ spaces into a single space."""
    return CONSECUTIVE_SPACES_REGEX.sub(" ", text)


@functools.lru_cache()
def get_blank_nlp(lang: str) -> spacy.Language:
    """Return a blank (without model) spaCy language pipeline."""
    return spacy.blank(lang)


@functools.lru_cache()
def get_lemmatizing_nlp(lang: str) -> spacy.Language:
    """Return a spaCy language pipeline with a lookup lemmatizer."""
    nlp = spacy.blank(lang)
    nlp.add_pipe("lemmatizer", config={"mode": "lookup"})
    nlp.initialize()
    return nlp


def get_tag(text: str) -> str:
    """Return a tag from a text.

    In Open Food Facts, tags are obtained from free text by performing the
    following:
    - lowercasing
    - accent removal
    - replacement of punctuation by either a comma ("-") or nothing, depending
    on the punctuation
    """
    text = strip_accents_ascii_v2(text)
    text = (
        text.lower()
        .replace(" & ", "-")
        .replace(" ", "-")
        .replace("'", "-")
        .replace(".", "-")
        .replace("!", "")
        .replace("?", "")
    )
    return strip_consecutive_spaces(text).strip("-")



import re
from typing import Union

from robotoff import settings
from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType
from robotoff.utils import text_file_iter

from .dataclass import OCRField, OCRRegex, OCRResult, get_text


def get_store_tag(store: str) -> str:
    return store.lower().replace(" & ", "-").replace(" ", "-").replace("'", "-")


def store_sort_key(item):
    """Sorting function for STORE_DATA items.
    For the regex to work correctly, we want the longest store names to
    appear first.
    """
    store, _ = item

    return -len(store), store


def get_sorted_stores() -> list[tuple[str, str]]:
    sorted_stores: dict[str, str] = {}

    for item in text_file_iter(settings.OCR_STORES_DATA_PATH):
        if "||" in item:
            store, regex_str = item.split("||")
        else:
            store = item
            regex_str = re.escape(item.lower())

        sorted_stores[store] = regex_str

    return sorted(sorted_stores.items(), key=store_sort_key)


SORTED_STORES = get_sorted_stores()
STORE_REGEX_STR = "|".join(
    r"((?<!\w){}(?!\w))".format(pattern) for _, pattern in SORTED_STORES
)
NOTIFY_STORES: set[str] = set(text_file_iter(settings.OCR_STORES_NOTIFY_DATA_PATH))
STORE_REGEX = OCRRegex(
    re.compile(STORE_REGEX_STR), field=OCRField.full_text_contiguous, lowercase=True
)


def find_stores(content: Union[OCRResult, str]) -> list[Prediction]:
    results = []

    text = get_text(content, STORE_REGEX)

    if not text:
        return []

    for match in STORE_REGEX.regex.finditer(text):
        groups = match.groups()

        for idx, match_str in enumerate(groups):
            if match_str is not None:
                store, _ = SORTED_STORES[idx]
                results.append(
                    Prediction(
                        type=PredictionType.store,
                        value=store,
                        value_tag=get_store_tag(store),
                        data={"text": match_str, "notify": store in NOTIFY_STORES},
                        predictor="regex",
                    )
                )
                break

    return results
